#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Signal and Port Handling.

A [Signal][ucdp.signal.Signal] is a module internal signal.
A [Port][ucdp.signal.Port] is a module interface signal with a direction IN, OUT or INOUT.

??? Example "Simple Types"
    Usage:

        >>> from tabulate import tabulate
        >>> import ucdp as u
        >>> u.Signal(u.UintType(6), "sig_s")
        Signal(UintType(6), 'sig_s')

        >>> u.Port(u.UintType(6), "port_i")
        Port(UintType(6), 'port_i', direction=IN)
        >>> u.Port(u.UintType(6), "port_o")
        Port(UintType(6), 'port_o', direction=OUT)

??? Example "Complex Types"
    Also complex types can simply be used:

        >>> class AType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("req", u.BitType())
        ...         self._add("data", u.ArrayType(u.UintType(16), 5))
        ...         self._add("ack", u.BitType(), u.BWD)
        >>> class MType(u.AEnumType):
        ...     keytype: u.AScalarType = u.UintType(2)
        ...     def _build(self) -> None:
        ...         self._add(0, "Linear")
        ...         self._add(1, "Cyclic")
        >>> class BType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("foo", AType())
        ...         self._add("mode", MType())
        ...         self._add("bar", u.ArrayType(AType(), 3), u.BWD)


??? Example "Resolving Types"
    These types are automatically resolved by iterating over the port:

        >>> sig = u.Signal(BType(), "sig_s")
        >>> for item in sig:
        ...     print(repr(item))
        Signal(BType(), 'sig_s')
        Signal(AType(), 'sig_foo_s')
        Signal(BitType(), 'sig_foo_req_s')
        Signal(ArrayType(UintType(16), 5), 'sig_foo_data_s')
        Signal(BitType(), 'sig_foo_ack_s', direction=BWD)
        Signal(MType(), 'sig_mode_s')
        Signal(ArrayType(AType(), 3), 'sig_bar_s', direction=BWD)
        Signal(ArrayType(BitType(), 3), 'sig_bar_ack_s')
        Signal(ArrayType(ArrayType(UintType(16), 5), 3), 'sig_bar_data_s', direction=BWD)
        Signal(ArrayType(BitType(), 3), 'sig_bar_req_s', direction=BWD)

        >>> inp = u.Port(BType(), "inp_i")
        >>> for item in inp:
        ...     print(repr(item))
        Port(BType(), 'inp_i', direction=IN)
        Port(AType(), 'inp_foo_i', direction=IN)
        Port(BitType(), 'inp_foo_req_i', direction=IN)
        Port(ArrayType(UintType(16), 5), 'inp_foo_data_i', direction=IN)
        Port(BitType(), 'inp_foo_ack_o', direction=OUT)
        Port(MType(), 'inp_mode_i', direction=IN)
        Port(ArrayType(AType(), 3), 'inp_bar_o', direction=OUT)
        Port(ArrayType(BitType(), 3), 'inp_bar_ack_i', direction=IN)
        Port(ArrayType(ArrayType(UintType(16), 5), 3), 'inp_bar_data_o', direction=OUT)
        Port(ArrayType(BitType(), 3), 'inp_bar_req_o', direction=OUT)

"""

from typing import ClassVar

from .casting import Casting
from .clkrelbase import BaseClkRel
from .ident import Ident
from .nameutil import split_suffix
from .object import PosArgs
from .orientation import DIRECTION_SUFFIXES, FWD, AOrientation, Direction
from .typebase import BaseType
from .typestruct import StructItem


class BaseSignal(Ident):
    """Base Class for All Signals ([Port][ucdp.signal.Port], [Signal][ucdp.signal.Signal]).

    Args:
        type_: Type.
        name: Name.

    Attributes:
        direction: Direction.
        doc: Documentation Container
        ifdef: IFDEF encapsulation

    """

    direction: AOrientation = FWD
    clkrel: BaseClkRel | None = None

    def cast(self, other: Ident) -> Casting:
        """Cast self=cast(other)."""
        return self.type_.cast(other.type_)

    def _new_structitem(self, structitem: StructItem, **kwargs) -> "BaseSignal":
        # self.clkrel.clk should be a true signal at this time
        clkrel = structitem.clkrel
        return super()._new_structitem(structitem, clkrel=clkrel or self.clkrel, **kwargs)


class Signal(BaseSignal):
    """
    Module Internal Signal.

    Args:
        type_: Type.
        name: Name.

    Attributes:
        direction: Direction.
        doc: Documentation Container
        ifdef: IFDEF encapsulation

    ??? Example "Signal Examples"
        Examples.

            >>> import ucdp as u
            >>> count = u.Signal(u.UintType(6), "count_s")
            >>> count
            Signal(UintType(6), 'count_s')
            >>> count.type_
            UintType(6)
            >>> count.name
            'count_s'
            >>> count.basename
            'count'
            >>> count.suffix
            '_s'
            >>> count.direction
            FWD
            >>> count.ifdefs
            ()

    Note:
        Signal names should end with '_r' or '_s'.
    """

    def __init__(self, type_: BaseType, name: str, **kwargs):
        kwargs.setdefault("direction", FWD)
        super().__init__(type_=type_, name=name, **kwargs)  # type: ignore[call-arg]


class Port(BaseSignal):
    """
    Module Port.

    Args:
        type_: Type.
        name: Name.

    Attributes:
        direction: Direction.
        doc: Documentation Container
        ifdef: IFDEF encapsulation

    ??? Example "Port Examples"
        Examples.

            >>> import ucdp as u
            >>> count = u.Port(u.UintType(6), "count_i")
            >>> count
            Port(UintType(6), 'count_i', direction=IN)
            >>> count.type_
            UintType(6)
            >>> count.name
            'count_i'
            >>> count.basename
            'count'
            >>> count.suffix
            '_i'
            >>> count.ifdefs
            ()

    ??? Example "Direction"
        The port direction is automatically determined from the name or needs to be specified explicitly:

            >>> u.Port(u.UintType(6), "count_i").direction
            IN
            >>> u.Port(u.UintType(6), "count_o").direction
            OUT
            >>> u.Port(u.UintType(6), "count_io").direction
            INOUT
            >>> u.Port(u.UintType(6), "count").direction
            Traceback (most recent call last):
            ...
            ValueError: 'direction' is required (could not be retrieved from name 'count').
            >>> u.Port(u.UintType(6), "count", direction=u.OUT).direction
            OUT
    """

    direction: Direction
    _posargs: ClassVar[PosArgs] = ("type_", "name")

    def __init__(self, type_: BaseType, name: str, **kwargs):
        if "direction" not in kwargs:
            kwargs["direction"] = direction = Direction.from_name(name)
            if direction is None:
                raise ValueError(f"'direction' is required (could not be retrieved from name {name!r}).")
        super().__init__(type_=type_, name=name, **kwargs)  # type: ignore[call-arg]

    @property
    def basename(self):
        """Base Name."""
        return split_suffix(self.name, only=DIRECTION_SUFFIXES)[0]

    @property
    def suffix(self):
        """Suffix."""
        return split_suffix(self.name, only=DIRECTION_SUFFIXES)[1]



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Constants."""

import re
import sys
from pathlib import Path
from typing import Literal

AUTO: str = "auto"
"""AUTO."""

PAT_IDENTIFIER: str = r"^[a-zA-Z]([a-zA-Z_0-9]*[a-zA-Z0-9])?$"
"""Pattern for Identifier."""

RE_IDENTIFIER = re.compile(PAT_IDENTIFIER)
"""Regular Expression for Identifier."""

PAT_OPT_IDENTIFIER: str = r"^([a-zA-Z]([a-zA-Z_0-9]*[a-zA-Z0-9])?)?$"
"""Pattern for Optional Identifier."""

PAT_IDENTIFIER_LOWER: str = r"^[a-z]([a-z_0-9]*[a-z0-9])?$"
"""Pattern for Identifier."""

PAT_IFDEF: str = r"^!?[a-zA-Z]([a-zA-Z_0-9]*[a-zA-Z0-9])?$"
"""Pattern for IFDEF."""

RE_IFDEF = re.compile(PAT_IFDEF)
"""Regular Expression for IFDEF."""

UPWARDS: str = ".."
"""UPWARDS."""

Gen = Literal["no", "inplace", "full", "custom"]
"""Gen."""

PATH = Path(__file__).parent

PKG_PATHS = {sys.prefix, sys.base_prefix}

TEMPLATE_PATHS = ("ucdp-templates/",)
"""Search Directories for Templates."""

CLINAME: str = "ucdp"
"""Name of Command-Line-Script - Used for documentation purposes."""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Expression Resolver.
"""

from typing import Any, ClassVar

from .expr import (
    BoolOp,
    ConcatExpr,
    ConstExpr,
    Expr,
    Log2Expr,
    MaximumExpr,
    MinimumExpr,
    Op,
    RangeExpr,
    SliceOp,
    SOp,
    TernaryExpr,
)
from .ident import Ident, Idents
from .namespace import Namespace
from .note import Note
from .object import Object
from .slices import Slice
from .typearray import ArrayType
from .typebase import BaseScalarType, BaseType
from .typeenum import BaseEnumType
from .typefloat import DoubleType, FloatType
from .typescalar import BitType, BoolType, IntegerType, RailType, SintType, UintType
from .typestring import StringType


class ExprResolver(Object):
    """
    Expression Resolver.

    ??? Example "Maximum Parser Example"
        Basics:

            >>> import ucdp as u
            >>> idents = u.Idents([
            ...     u.Signal(u.UintType(16), 'uint_s'),
            ...     u.Signal(u.SintType(16), 'sint_s'),
            ... ])
            >>> parser = u.ExprParser(namespace=idents)
            >>> expr = parser.parse('uint_s') * parser.const(2)
            >>> expr
            Op(Signal(UintType(16), 'uint_s'), '*', ConstExpr(IntegerType(default=2)))

            >>> resolver = u.ExprResolver(namespace=idents)
            >>> resolver.resolve(expr)
            'uint_s * 2'
            >>> resolver.resolve(expr, brackets=True)
            '(uint_s * 2)'
    """

    namespace: Namespace | None = None
    remap: Idents | None = None

    _opremap: ClassVar[dict[str, str]] = {}
    _BRACKETTYPES: tuple[Any, ...] = (Op, BoolOp, SOp, TernaryExpr)

    def __call__(self, expr: Expr, brackets: bool = False) -> str:
        """
        Resolve.

        Args:
            expr: Expression
            brackets: Use brackets if necessary for topmost expr.
        """
        return self._resolve(expr, brackets=brackets)

    def resolve(self, expr: Expr | Note, brackets: bool = False) -> str:
        """
        Resolve.

        Args:
            expr: Expression
            brackets: Use brackets if necessary for topmost expr.
        """
        return self._resolve(expr, brackets=brackets)

    def _resolve(self, expr: Expr | Note, brackets: bool = False) -> str:  # noqa: C901, PLR0912
        if isinstance(expr, Ident):
            resolved = self._resolve_ident(expr)
        elif isinstance(expr, BoolOp):
            resolved = self._resolve_boolop(expr)
        elif isinstance(expr, SOp):
            resolved = self._resolve_sop(expr)
        elif isinstance(expr, Op):
            resolved = self._resolve_op(expr)
        elif isinstance(expr, SliceOp):
            resolved = self._resolve_sliceop(expr)
        elif isinstance(expr, ConstExpr):
            resolved = self._resolve_constexpr(expr)
        elif isinstance(expr, ConcatExpr):
            resolved = self._resolve_concatexpr(expr)
        elif isinstance(expr, TernaryExpr):
            resolved = self._resolve_ternaryexpr(expr)
        elif isinstance(expr, Log2Expr):
            resolved = self._resolve_log2expr(expr)
        elif isinstance(expr, MinimumExpr):
            resolved = self._resolve_minimumexpr(expr)
        elif isinstance(expr, MaximumExpr):
            resolved = self._resolve_maximumexpr(expr)
        elif isinstance(expr, RangeExpr):
            resolved = self._resolve_rangeexpr(expr)
        elif isinstance(expr, Note):
            resolved = self._get_note(expr)
        else:
            raise ValueError(f"{expr!r} is not a valid expression.")
        if brackets and isinstance(expr, self._BRACKETTYPES):
            resolved = f"({resolved})"
        return resolved

    def _resolve_ident(self, ident: Ident) -> str:
        # Remapping of identifier, i.e. on instance port list
        if self.remap is not None:
            if ident.name in self.remap.keys():
                ref = self.remap[ident.name]
                if ref.value is not None and ref.value != ref:
                    # resolve remappend identifier value
                    return self.resolve(ref.value)
                # just use default value
                return self._resolve_value(ident.type_)

        # Namespace checking
        if self.namespace is not None:
            # check if identifier exists in namespace.
            if ident.name not in self.namespace:
                raise ValueError(f"{ident!r} not known within current namespace.")

        return ident.name

    def _resolve_op(self, op: Op) -> str:
        left = self._resolve(op.left, brackets=True)
        right = self._resolve(op.right, brackets=True)
        sign = self._opremap.get(op.sign, op.sign)
        return f"{left} {sign} {right}"

    def _resolve_boolop(self, op: BoolOp) -> str:
        left = self._resolve(op.left, brackets=True)
        right = self._resolve(op.right, brackets=True)
        return f"{left} {op.sign} {right}"

    def _resolve_sop(self, op: SOp) -> str:
        one = self._resolve(op.one)
        return f"{op.sign}{one}{op.postsign}"

    def _resolve_sliceop(self, op: SliceOp) -> str:
        one = self._resolve(op.one)
        return f"{one}{self._resolve_slice(op.slice_, opt=True)}"

    def resolve_slice(self, slice_: Slice, opt: bool = True) -> str:
        """Resolve Slice."""
        return self._resolve_slice(slice_, opt=opt)

    def _resolve_slice(self, slice_: Slice, opt: bool = False) -> str:
        left = slice_.left
        right = slice_.right
        if opt and left is right:
            return f"[{left}]"

        if not isinstance(left, int):
            left = self.resolve(left)
        if not isinstance(right, int):
            right = self.resolve(right)
        return f"[{left}:{right}]"

    def _resolve_concatexpr(self, expr: ConcatExpr) -> str:
        items = ", ".join(self._resolve(item) for item in expr.items)
        return f"{{{items}}}"

    def _resolve_ternaryexpr(self, expr: TernaryExpr) -> str:
        cond = self._resolve(expr.cond, brackets=True)
        one = self._resolve(expr.one, brackets=True)
        other = self._resolve(expr.other, brackets=True)
        return f"{cond} ? {one} : {other}"

    def _resolve_log2expr(self, expr: Log2Expr) -> str:
        raise NotImplementedError

    def _resolve_minimumexpr(self, expr: MinimumExpr) -> str:
        raise NotImplementedError

    def _resolve_maximumexpr(self, expr: MaximumExpr) -> str:
        raise NotImplementedError

    def _resolve_rangeexpr(self, expr: RangeExpr) -> str:
        raise NotImplementedError

    def _resolve_constexpr(self, expr: ConstExpr) -> str:
        try:
            return self._resolve_value(expr.type_)
        except ValueError as exc:
            raise ValueError(f"{expr} {exc}") from None

    def resolve_value(self, type_: BaseType, value=None) -> str:
        """Resolve Value."""
        return self._resolve_value(type_, value=value)

    def _resolve_value(self, type_: BaseType, value=None) -> str:  # noqa: C901, PLR0911, PLR0912
        if isinstance(type_, ArrayType):
            # TODO: value
            itemvalue = self._resolve_value(type_.itemtype)
            return self._get_array_value(itemvalue, type_.slice_)

        if not isinstance(type_, (BaseScalarType, StringType, FloatType, DoubleType)):
            raise ValueError(f"Cannot resolve type {type_}")
        if value is None:
            value = type_.default

        if isinstance(type_, StringType):
            return self._get_string_value(value)

        # None
        if value is None:
            return ""

        # Expr
        if isinstance(value, Expr):
            return self.resolve(value)

        while isinstance(type_, BaseEnumType):
            type_ = type_.keytype

        if isinstance(type_, BitType):
            return self._get_bit_value(int(value))

        if isinstance(type_, UintType):
            width = int(type_.width)
            if width < 1:
                raise ValueError(f"Invalid width {width}")
            return self._get_uint_value(value, type_.width)

        if isinstance(type_, SintType):
            width = int(type_.width)
            if width < 1:
                raise ValueError(f"Invalid width {width}")
            return self._get_sint_value(value, type_.width)

        if isinstance(type_, IntegerType):
            return self._get_integer_value(value)

        if isinstance(type_, RailType):
            return self._get_rail_value(value)

        if isinstance(type_, BoolType):
            return self._get_bool_value(value)

        if isinstance(type_, FloatType):
            return self._get_float_value(value)

        if isinstance(type_, DoubleType):
            return self._get_double_value(value)

        raise AssertionError

    @staticmethod
    def _get_rail_value(value: int) -> str:
        return str(value)

    @staticmethod
    def _get_bit_value(value: int) -> str:
        return str(value)

    @staticmethod
    def _get_uint_value(value: int, width: int | Expr) -> str:
        return f"0x{value:X}"

    @staticmethod
    def _get_sint_value(value: int, width: int | Expr) -> str:
        if value < 0:
            return f"-0x{-value:X}"
        return f"0x{value:X}"

    @staticmethod
    def _get_integer_value(value: int) -> str:
        return str(value)

    @staticmethod
    def _get_bool_value(value: bool) -> str:
        return str(value)

    @staticmethod
    def _get_string_value(value: int) -> str:
        return repr(value)

    @staticmethod
    def _get_note(note: Note) -> str:
        return repr(note.note)

    def _get_array_value(self, itemvalue: str, slice_: Slice) -> str:
        raise NotImplementedError

    @staticmethod
    def _get_float_value(value: float) -> str:
        return f"{value:.1f}"

    @staticmethod
    def _get_double_value(value: float) -> str:
        return f"{value:.1f}"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Utilities.
"""

import functools
import re
from collections.abc import Iterable

from matchor import matchs


@functools.lru_cache(maxsize=128)
def _translate(seps):
    return re.compile(rf"[{seps}]\s*")


Names = Iterable[str] | str


def split(items: Names, seps: str = ";") -> tuple[str, ...]:
    """
    Split items into tuple.

    Args:
        items: Items.
        seps: Separator

    ??? Example "Split Examples"
        Basics:

            >>> split(['abc', 'def'])
            ('abc', 'def')
            >>> split(('abc', 'def'))
            ('abc', 'def')
            >>> split('abc; def')
            ('abc', 'def')
            >>> split('ab;c, def', seps=",")
            ('ab;c', 'def')

        Generators are also handled:

            >>> def func():
            ...     yield 'abc'
            ...     yield 'def'
            >>> split(func())
            ('abc', 'def')

        Empty strings or `None` just result in an empty tuple:

            >>> split("")
            ()
            >>> split(None)
            ()
    """
    if not items:
        items = []
    elif isinstance(items, str):
        items = [item.strip() for item in _translate(seps).split(items)]
    return tuple(items)


def namefilter(namepats: Names):
    """
    Create filter for namepats.

    Args:
        namepats: Name Filter Pattern.

    ??? Example "Namefilter Examples"
        Basics:

            >>> import ucdp as u
            >>> def myfunc(items, filter_=None):
            ...     for item in items:
            ...         if not filter_ or filter_(item):
            ...             print(item)

            >>> items = ('abc', 'cde', 'efg')
            >>> myfunc(items)
            abc
            cde
            efg
            >>> myfunc(items, filter_=u.namefilter(''))
            abc
            cde
            efg
            >>> myfunc(items, filter_=u.namefilter('cde; tuv'))
            cde
            >>> myfunc(items, filter_=u.namefilter('*c*'))
            abc
            cde
            >>> myfunc(items, filter_=u.namefilter('!*c*'))
            efg
            >>> myfunc(items, filter_=u.namefilter('*c*; !*e'))
            abc

            >>> items = ['abc', 'cde', 'efg']
            >>> myfunc(items, filter_=u.namefilter('*c*; !*e'))
            abc
    """
    _patterns: tuple[str, ...] = split(namepats)

    inclist = [pat for pat in _patterns if not pat.startswith("!")]
    exclist = [pat[1:] for pat in _patterns if pat.startswith("!")]

    if inclist:
        if exclist:

            def filter_(name: str):
                return matchs(name, inclist) and not matchs(name, exclist)
        else:

            def filter_(name: str):
                return matchs(name, inclist)
    elif exclist:

        def filter_(name: str):
            return not matchs(name, exclist)
    else:

        def filter_(name: str):
            return True

    return filter_



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Identifier."""

from pytest import fixture

import ucdp as u


class ModeType(u.AEnumType):
    """Mode."""

    keytype: u.UintType = u.UintType(2)

    def _build(self):
        self._add(0, "add")
        self._add(1, "sub")
        self._add(2, "max")


class StructType(u.AStructType):
    """My."""

    comment: str = "Mode"

    def _build(self):
        self._add("mode", ModeType())
        self._add("send", u.ArrayType(u.UintType(8), 3))
        self._add("return", u.UintType(4), u.BWD)


class MyType(u.AStructType):
    """A Complex Type."""

    def _build(self):
        self._add("my0", StructType())
        self._add("my1", StructType(), u.BWD)
        self._add("uint", u.UintType(3))


@fixture
def top() -> u.Idents:
    """Top-Identifier."""
    return u.Idents(
        [
            u.Port(MyType(), "port_i"),
            u.Port(MyType(), "other_o"),
            u.Port(StructType(), "struct_i"),
            u.Port(StructType(), "struct_o"),
            u.Signal(StructType(), "struct_s"),
            u.Signal(MyType(), "sig_s"),
            u.Port(u.UintType(8), "data_i"),
        ]
    )


def test_expridents():
    """Expression Identifier Resolve."""
    ident0 = u.Ident(u.UintType(8), "ident0")
    ident1 = u.Ident(u.UintType(8), "ident1")
    ident2 = u.Ident(u.UintType(8), "ident2")

    assert u.get_expridents(ident0 + -ident1) == (ident0, ident1)
    assert u.get_expridents(ident0 + 5) == (ident0,)
    assert u.get_expridents(ident0 + 5) == (ident0,)
    assert u.get_expridents(u.ConcatExpr((ident0, ident1))) == (ident0, ident1)
    assert u.get_expridents(u.Log2Expr(ident0)) == (ident0,)

    assert u.get_expridents(u.TernaryExpr(ident2 == 5, ident1, ident0)) == (ident2, ident1, ident0)


def test_idents():
    """Idents."""
    port = u.Port(u.ClkRstAnType(), "main_i")
    idents = u.Idents([port])
    assert idents["main_i"] is port
    assert idents["main_i"] == port
    assert "main_clk_i" in idents
    assert idents["main_clk_i"] == u.Port(u.ClkType(), "main_clk_i", direction=u.IN, doc=u.Doc(title="Clock"))


def test_leveliter(top):
    """Top."""
    assert tuple(f"{idx}: {name}" for idx, name in top.leveliter()) == (
        "0: port_i",
        "1: port_my0_i",
        "2: port_my0_mode_i",
        "2: port_my0_send_i",
        "2: port_my0_return_o",
        "1: port_my1_o",
        "2: port_my1_mode_o",
        "2: port_my1_send_o",
        "2: port_my1_return_i",
        "1: port_uint_i",
        "0: other_o",
        "1: other_my0_o",
        "2: other_my0_mode_o",
        "2: other_my0_send_o",
        "2: other_my0_return_i",
        "1: other_my1_i",
        "2: other_my1_mode_i",
        "2: other_my1_send_i",
        "2: other_my1_return_o",
        "1: other_uint_o",
        "0: struct_i",
        "1: struct_mode_i",
        "1: struct_send_i",
        "1: struct_return_o",
        "0: struct_o",
        "1: struct_mode_o",
        "1: struct_send_o",
        "1: struct_return_i",
        "0: struct_s",
        "1: struct_mode_s",
        "1: struct_send_s",
        "1: struct_return_s",
        "0: sig_s",
        "1: sig_my0_s",
        "2: sig_my0_mode_s",
        "2: sig_my0_send_s",
        "2: sig_my0_return_s",
        "1: sig_my1_s",
        "2: sig_my1_mode_s",
        "2: sig_my1_send_s",
        "2: sig_my1_return_s",
        "1: sig_uint_s",
        "0: data_i",
    )


def test_port_iter():
    """Port Iteration."""
    assert tuple(sub.name for sub in u.Port(MyType(), "port_i").iter()) == (
        "port_i",
        "port_my0_i",
        "port_my0_mode_i",
        "port_my0_send_i",
        "port_my0_return_o",
        "port_my1_o",
        "port_my1_mode_o",
        "port_my1_send_o",
        "port_my1_return_i",
        "port_uint_i",
    )
    assert tuple(sub.name for sub in u.Port(MyType(), "port", direction=u.IN).iter()) == (
        "port",
        "port_my0",
        "port_my0_mode",
        "port_my0_send",
        "port_my0_return",
        "port_my1",
        "port_my1_mode",
        "port_my1_send",
        "port_my1_return",
        "port_uint",
    )
    assert tuple(sub.name for sub in u.Port(MyType(), "po_t", direction=u.IN).iter()) == (
        "po_t",
        "po_t_my0",
        "po_t_my0_mode",
        "po_t_my0_send",
        "po_t_my0_return",
        "po_t_my1",
        "po_t_my1_mode",
        "po_t_my1_send",
        "po_t_my1_return",
        "po_t_uint",
    )



"""My Name No Regf Module."""


from fileliststandard import HdlFileList
from glbl_lib.bus import BusType

import ucdp as u


class MyNameNoRegfIoType(u.AStructType):
    """My Name No Regf IO."""

    title: str = "My Name No Regf"
    comment: str = "RX/TX"

    def _build(self) -> None:
        self._add("rx", u.BitType(), u.BWD)
        self._add("tx", u.BitType(), u.FWD)


class MyNameNoRegfMod(u.AMod):
    """My Name No Regf Module."""

    filelists: u.ClassVar[u.ModFileLists] = (
        HdlFileList(gen="full"),
    )

    def _build(self) -> None:
        """Build."""
        self.add_port(u.ClkRstAnType(), "main_i")
        self.add_port(MyNameNoRegfIoType(), "my_name_no_regf_i", clkrel=u.ASYNC)
        self.add_port(BusType(), "bus_i", clkrel="main_clk_i")




#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Loader and Top."""

import ucdp as u


def test_load_simple(example_simple):
    """Simple Module."""
    top = u.load("glbl_lib.clk_gate", paths=None)
    assert u.TopModRef.from_mod(top.mod) == top.ref


def test_load_complex(example_simple):
    """Complexer Module."""
    top = u.load("uart_lib.uart", paths=None)
    assert u.TopModRef.from_mod(top.mod) == top.ref
    assert u.TopModRef.from_mod(top.mod.get_inst("u_clk_gate")) == u.TopModRef(u.ModRef("glbl_lib", "clk_gate"))
    assert u.TopModRef.from_mod(top.mod.get_inst("u_core")) == u.TopModRef(
        u.ModRef("uart_lib", "uart"), sub="uart_lib.uart_core"
    )
    assert u.TopModRef.from_mod(top.mod.get_inst("u_regf")) == u.TopModRef(
        u.ModRef("uart_lib", "uart"), sub="uart_lib.uart_regf"
    )


def test_tb(example_simple):
    """Testbench."""
    top = u.load("glbl_lib.regf_tb#uart_lib.uart-uart_lib.uart_regf", paths=None)
    assert u.TopModRef.from_mod(top.mod) == top.ref



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Configuration."""

from pytest import raises
from uniquer import uniquelist

import ucdp as u


class MyConfig(u.AConfig):
    """My Configuration."""

    _hash_excludes: u.ClassVar[set[str]] = {
        "ignored",
    }

    mem_baseaddr: u.Hex
    ram_size: u.Bytesize
    rom_size: u.Bytesize = 0
    feature: bool = False

    ignored: int = 0


def test_config():
    """Example."""
    # Missing Arguments
    with raises(u.ValidationError):
        MyConfig(name="myconfig")

    config = MyConfig("myconfig", mem_baseaddr=0xF100, ram_size="16 kB")
    assert str(config) == "MyConfig('myconfig', mem_baseaddr=Hex('0xF100'), ram_size=Bytesize('16 KB'))"
    assert dict(config) == {
        "feature": False,
        "ignored": 0,
        "mem_baseaddr": u.Hex("0xF100"),
        "modname": "",
        "name": "myconfig",
        "ram_size": u.Bytesize("16 KB"),
        "rom_size": u.Bytesize("0 bytes"),
    }

    assert config.hash == "5f9de9502e167f99"
    assert config.is_default is False


class OtherConfig(u.AConfig):
    """Other Configuration."""

    ram_size: u.Bytesize = 0x100
    rom_size: u.Bytesize = 0
    feature: bool = False


def test_default_config():
    """Default Configuration."""
    config = OtherConfig()
    assert config.hash == "286f2a5127a1bd37"
    assert config.is_default is True

    config = OtherConfig(name="abc")
    assert config.hash == "619017c580520b19"
    assert config.is_default is False


class ParentConfig(u.AConfig):
    """My Configuration."""

    _hash_excludes: u.ClassVar[set[str]] = {
        "ignored",
    }

    mem_baseaddr: u.Hex
    sub: MyConfig
    subs: tuple[MyConfig, ...]

    ignored: int = 0


def assert_unique_hashes(*configs):
    """Check for unique hashes."""
    hashes = [config.hash for config in configs]
    assert hashes == uniquelist(hashes)


def test_hier_config():
    """Test hierarchical hashes."""
    my0 = MyConfig(mem_baseaddr=0x1000, ram_size=0x2000)
    my1 = MyConfig(mem_baseaddr=0x1000, ram_size=0x2001)
    my2 = MyConfig(mem_baseaddr=0x1000, ram_size=0x2001, ignored=1)
    assert my0.hash != my1.hash
    assert my1.hash == my2.hash

    # ignored in parent
    p0 = ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my0, my0))
    p1 = ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my0, my0), ignored=1)
    assert p0.hash == p1.hash

    # changed sub
    p2 = ParentConfig(mem_baseaddr=0x1000, sub=my1, subs=(my0, my0))
    assert p0.hash != p2.hash

    # ignored in sub
    p3 = ParentConfig(mem_baseaddr=0x1000, sub=my2, subs=(my0, my0))
    assert p2.hash == p3.hash

    # changed in subs
    assert_unique_hashes(
        ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my0, my0)),
        ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my1, my0)),
        ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my0, my1)),
        ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my1, my1)),
    )

    # ignored in subs
    p4 = ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my1, my0))
    p5 = ParentConfig(mem_baseaddr=0x1000, sub=my0, subs=(my2, my0))
    assert p4.hash == p5.hash



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Testbench Module.
"""

from collections.abc import Sequence
from typing import ClassVar

from .modbase import BaseMod
from .modbasetop import BaseTopMod
from .modfilelist import ModFileLists
from .modutil import get_libpath, get_modname, get_topmodname
from .object import Field
from .test import Test


class ATbMod(BaseTopMod):
    """
    Testbench Module.

    Attributes:
        filelists: File Lists.
        dut_modclss: Testbench is limited to these kind of modules.
        title: Title.
        dut: Module Under Test.
        parent: Parent.
    """

    filelists: ClassVar[ModFileLists] = ()
    """File Lists."""

    title: str = "Testbench"

    parent: BaseMod | None = Field(default=None, init=False)

    @property
    def modname(self) -> str:
        """Module Name."""
        return get_modname(self.__class__)

    @property
    def topmodname(self) -> str:
        """Top Module Name."""
        return get_topmodname(self.__class__)

    @property
    def libpath(self) -> str:
        """Library Path."""
        return get_libpath(self.__class__)

    @property
    def is_tb(self) -> bool:
        """Determine if module belongs to Testbench or Design."""
        return True

    def get_tests(self) -> Sequence[Test]:
        """Get Tests."""
        return ()



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Clock and Reset Types.

* :any:`ClkType` - Clock
* :any:`RstAType` - Asynchronous Reset (High-Active)
* :any:`RstAnType` - Asynchronous Reset (Low-Active)
* :any:`RstType` - Synchronous Reset (High-Active)
* :any:`ClkRstAnType` - Clock and Async Reset Pair
* :any:`DiffClkType` - Differential Clock
* :any:`DiffClkRstAnType` - Differential Clock and Async Reset
"""

from .clkrelbase import BaseClkRel
from .typebase import AScalarType
from .typeenum import AEnumType
from .typescalar import BitType
from .typestruct import AStructType


class ClkType(BitType):
    """
    Clock Type.

    Single bit used as clock.

    >>> import ucdp as u
    >>> clk = u.ClkType()
    >>> clk
    ClkType()
    >>> clk.width
    1
    """

    title: str = "Clock"


class RstAType(AEnumType):
    """
    Async high-active Reset Type.

    Single bit used as asynchronous reset.

    >>> import ucdp as u
    >>> rst = u.RstAType()
    >>> rst
    RstAType()
    >>> rst.width
    1
    >>> for item in rst.values():
    ...     print(repr(item))
    EnumItem(0, 'inactive', doc=Doc(title='Reset not active'))
    EnumItem(1, 'active', doc=Doc(title='Reset active'))
    """

    keytype: AScalarType = BitType()
    title: str = "Async Reset"
    descr: str = "High-Active"
    comment: str = "Async Reset (High-Active)"

    def _build(self) -> None:
        self._add(0, "inactive", "Reset not active")
        self._add(1, "active", "Reset active")


class RstAnType(AEnumType):
    """
    Async low-active Reset Type.

    Single bit used as asynchronous low-active reset.

    >>> import ucdp as u
    >>> rst = u.RstAnType()
    >>> rst
    RstAnType()
    >>> rst.width
    1
    >>> for item in rst.values():
    ...     print(repr(item))
    EnumItem(0, 'active', doc=Doc(title='Reset active'))
    EnumItem(1, 'inactive', doc=Doc(title='Reset not active'))
    """

    keytype: AScalarType = BitType()
    title: str = "Async Reset"
    descr: str = "Low-Active"
    comment: str = "Async Reset (Low-Active)"

    def _build(self) -> None:
        self._add(0, "active", "Reset active")
        self._add(1, "inactive", "Reset not active")


class RstType(AEnumType):
    """
    Sync Reset.

    Single bit used as synchronous reset.

    >>> import ucdp as u
    >>> rst = u.RstType()
    >>> rst
    RstType()
    >>> rst.width
    1
    >>> for item in rst.values():
    ...     print(repr(item))
    EnumItem(0, 'inactive', doc=Doc(title='Reset not active'))
    EnumItem(1, 'active', doc=Doc(title='Reset active'))
    """

    keytype: AScalarType = BitType()
    title: str = "Synchronous Reset"
    descr: str = "High-Active"

    def _build(self) -> None:
        self._add(0, "inactive", "Reset not active")
        self._add(1, "active", "Reset active")


class ClkRstAnType(AStructType):
    """
    Clock, Async Reset Pair.

    >>> import ucdp as u
    >>> clkrst = u.ClkRstAnType()
    >>> for item in clkrst.values():
    ...     print(repr(item))
    StructItem('clk', ClkType(), doc=Doc(title='Clock'))
    StructItem('rst_an', RstAnType(), doc=Doc(title='Async Reset', descr='Low-Active', comment='Async Reset ...'))
    """

    title: str = "Clock and Reset"
    clkrel: BaseClkRel | None = None

    def _build(self) -> None:
        self._add("clk", ClkType(), clkrel=self.clkrel)
        self._add("rst_an", RstAnType())


class DiffClkType(AStructType):
    """
    Differential Clock.

    >>> import ucdp as u
    >>> diffclk = u.DiffClkType()
    >>> for item in diffclk.values():
    ...     print(repr(item))
    StructItem('p', ClkType(), doc=Doc(title='Clock'))
    StructItem('n', ClkType(default=1), doc=Doc(title='Inverted Clock'))
    """

    title: str = "Differential Clock"
    clkrel: BaseClkRel | None = None

    def _build(self) -> None:
        self._add("p", ClkType(default=0), clkrel=self.clkrel)
        self._add("n", ClkType(default=1), title="Inverted Clock", clkrel=self.clkrel)


class DiffClkRstAnType(AStructType):
    """
    Differential Clock and Reset.

    >>> import ucdp as u
    >>> diffclkrst = u.DiffClkRstAnType()
    >>> for item in diffclkrst.values():
    ...     print(repr(item))
    StructItem('clk', DiffClkType(), doc=Doc(title='Differential Clock'))
    StructItem('rst_an', RstAnType(), doc=Doc(title='Async Reset', descr='Low-Active', comment='Async Reset ...'))
    """

    title: str = "Differential Clock and Reset"
    clkrel: BaseClkRel | None = None

    def _build(self) -> None:
        self._add("clk", DiffClkType(clkrel=self.clkrel))
        self._add("rst_an", RstAnType())



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Module Reference."""

# DOCME: pkg vs lib

import re
from typing import ClassVar, Union

from caseconverter import pascalcase

from .consts import PAT_IDENTIFIER
from .object import Field, Object, PosArgs

RE_MODREF = re.compile(
    # lib
    r"(?P<libname>[a-zA-Z][a-zA-Z_0-9]*)\."
    # mod
    r"(?P<modname>[a-zA-Z][a-zA-Z_0-9]*)"
    # cls
    r"(\.(?P<modclsname>[A-Z][a-zA-Z_0-9]*Mod))?"
)
PAT_MODREF = "lib.my[.MyMod]"


class ModRef(Object):
    """
    Module Reference.

    Args:
        libname: Library Name
        modname: Module Name

    Keyword Args:
        modclsname: Class Name

    ??? Example "Module Reference Examples"
        Examples.

            >>> ModRef('glbl_lib', 'clk_gate', modclsname='ClkGateMod')
            ModRef('glbl_lib', 'clk_gate', modclsname='ClkGateMod')

            Just a module:

            >>> spec = ModRef.cast('glbl_lib.clk_gate')
            >>> spec
            ModRef('glbl_lib', 'clk_gate')
            >>> str(spec)
            'glbl_lib.clk_gate'

            Module from a package and explicit class:

            >>> spec = ModRef.cast('glbl_lib.clk_gate.ClkGateMod')
            >>> spec
            ModRef('glbl_lib', 'clk_gate', modclsname='ClkGateMod')
            >>> str(spec)
            'glbl_lib.clk_gate.ClkGateMod'
            >>> ModRef.cast(ModRef('glbl_lib', 'clk_gate', modclsname='ClkGateMod'))
            ModRef('glbl_lib', 'clk_gate', modclsname='ClkGateMod')

            Invalid Pattern:

            >>> ModRef.cast('lib.my:c-ls')
            Traceback (most recent call last):
            ..
            ValueError: 'lib.my:c-ls' does not match pattern 'lib.my[.MyMod]'
    """

    libname: str = Field(pattern=PAT_IDENTIFIER)
    modname: str = Field(pattern=PAT_IDENTIFIER)
    modclsname: str | None = Field(pattern=PAT_IDENTIFIER, default=None)

    _posargs: ClassVar[PosArgs] = ("libname", "modname")

    def __init__(self, libname: str, modname: str, modclsname: str | None = None):
        super().__init__(libname=libname, modname=modname, modclsname=modclsname)  # type: ignore[call-arg]

    def __str__(self) -> str:
        modclsname = f".{self.modclsname}" if self.modclsname else ""
        return f"{self.libname}.{self.modname}{modclsname}"

    @staticmethod
    def cast(value: Union["ModRef", str]) -> "ModRef":
        """Cast `value` to `ModRef`."""
        if isinstance(value, ModRef):
            return value

        mat = RE_MODREF.fullmatch(value)
        if mat:
            return ModRef(**mat.groupdict())

        raise ValueError(f"{value!r} does not match pattern {PAT_MODREF!r}")

    def get_modclsname(self) -> str:
        """Return modclsname or derive from modname."""
        return self.modclsname or get_modclsname(self.modname)


def get_modclsname(modname: str) -> str:
    """Get Module Class Name."""
    return f"{pascalcase(modname)}Mod"



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Hardware Module Builder.

"""

from .exceptions import BuildError
from .logging import LOGGER
from .modbase import BaseMod
from .moditer import ModPreIter
from .nameutil import didyoumean

TIMEOUT: int = 10


def _stop_at_locked(mod: BaseMod):
    return mod.is_locked


def _build(mod: BaseMod, name: str):
    insts: set[int] = set()

    for iteration in range(1, TIMEOUT):
        LOGGER.debug("%s: builder: %s iteration %s", mod, name, iteration)
        pending = False
        for inst in ModPreIter(mod=mod, stop=_stop_at_locked):
            instid = hash((inst.qualname, inst.inst))

            # build every instance just once
            if instid in insts:
                continue
            insts.add(instid)

            # Determine build method
            try:
                build_method = getattr(inst, name)
            except AttributeError:
                continue
            if not getattr(inst, f"_has{name}"):
                raise TypeError(f"{inst} is not allowed to have a {name} method.")

            # Invoke build_method
            build_method()
            pending = True

        if not pending:
            break
    else:
        raise BuildError(f"Cannot build {mod}")


def build(mod: BaseMod):
    _build(mod, "_build_dep")

    _build(mod, "_build_final")

    insts = tuple(ModPreIter(mod, stop=_stop_at_locked))

    # route
    for inst in insts:
        inst._router.flush()

    # paramdict+lock
    for inst in insts:
        for name in inst.paramdict:
            advice = didyoumean(name, inst.params.keys(), known=True)
            raise ValueError(f"{inst} has no param {name!r}{advice}")
        # for mux in inst.muxes:
        #     for sel in mux.sels:
        #         if isinstance(sel.type_, BaseEnumType):
        #             inst.add_type_consts(sel.type_, exist_ok=True)
        inst.lock()



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Top Reference."""

import re
from pathlib import Path
from typing import ClassVar, Literal, Optional, Union

from pydantic_core import PydanticUndefined

from .mod import AMod
from .modbase import BaseMod, ModCls
from .modconfigurable import AConfigurableMod
from .modgenerictb import AGenericTbMod
from .modref import ModRef
from .modtb import ATbMod
from .object import Field, LightObject, PosArgs
from .pathutil import absolute

RE_TOPMODREF = re.compile(
    # [tb]#
    r"((?P<tb>[a-zA-Z_0-9_\.]+)#)?"
    # top
    r"(?P<top>[a-zA-Z_0-9_\.]+)"
    # [-sub]
    r"(-(?P<sub>[a-zA-Z_0-9_\.]+))?"
)
PAT_TOPMODREF = "[tb_lib.tb#]top_lib.top[-sub_lib.sub]"
RE_MODREF = r"[a-zA-Z][a-zA-Z_0-9]*\.[a-zA-Z][a-zA-Z_0-9]*"


TbType = Literal["Static", "Generic", ""]


class TopModRef(LightObject):
    """
    Top Module Reference.

    Args:
        top: Top Module Reference

    Attributes:
        sub: Sub Module Reference
        tb: Testbench Module Reference

    ??? Example "Top Reference Examples"
        Example:

            >>> import ucdp as u
            >>> u.TopModRef.cast('top_lib.top_mod')
            TopModRef(ModRef('top_lib', 'top_mod'))
            >>> u.TopModRef.cast('top_lib.top_mod-sub_lib.sub_mod')
            TopModRef(ModRef('top_lib', 'top_mod'), sub='sub_lib.sub_mod')
            >>> u.TopModRef.cast('mod_tb_lib.mod_tb#top_lib.top_mod')
            TopModRef(ModRef('top_lib', 'top_mod'), tb=ModRef('mod_tb_lib', 'mod_tb'))
            >>> u.TopModRef.cast(TopModRef(ModRef('top_lib', 'top_mod')))
            TopModRef(ModRef('top_lib', 'top_mod'))

            Invalid Pattern:

            >>> TopModRef.cast('lib.mod:c-ls.1')
            Traceback (most recent call last):
            ..
            ValueError: 'lib.mod:c-ls.1' does not match pattern '[tb_lib.tb#]top_lib.top[-sub_lib.sub]'
    """

    top: ModRef
    sub: str | None = Field(default=None, pattern=RE_MODREF)
    tb: ModRef | None = None

    _posargs: ClassVar[PosArgs] = ("top",)

    def __init__(self, top: ModRef, sub: str | None = None, tb: ModRef | None = None):
        super().__init__(top=top, sub=sub, tb=tb)  # type: ignore[call-arg]

    def __str__(self) -> str:
        result = str(self.top)
        if self.sub:
            result = f"{result}-{self.sub}"
        if self.tb:
            result = f"{self.tb}#{result}"
        return result

    @staticmethod
    def cast(value: Union["TopModRef", Path, str]) -> "TopModRef":
        """Cast `value` to `TopModRef`."""
        if isinstance(value, TopModRef):
            return value

        if isinstance(value, Path):
            path = absolute(value)
            modname = path.stem
            libname = path.parent.name
            return TopModRef(top=ModRef(libname=libname, modname=modname))

        mat = RE_TOPMODREF.fullmatch(value)
        if mat:
            top = ModRef.cast(mat.group("top"))
            sub = mat.group("sub")
            tb = ModRef.cast(mat.group("tb")) if mat.group("tb") else None
            return TopModRef(top=top, sub=sub, tb=tb)

        raise ValueError(f"{value!r} does not match pattern {PAT_TOPMODREF!r}")

    @staticmethod
    def from_mod(mod: BaseMod) -> Optional["TopModRef"]:
        """From Module."""
        if get_tb(mod.__class__) == "Generic":
            tbref = mod.get_modref(minimal=True)
            mod = mod.dut
        else:
            tbref = None
        if not is_top(mod.__class__):
            sub = f"{mod.libname}.{mod.modname}"
            while mod is not None:
                if is_top(mod.__class__):
                    break
                if isinstance(mod, AConfigurableMod):
                    # no standalone configuration
                    mod = None
                    break
                mod = mod.parent
        else:
            sub = None
        if mod is None:
            return None
        modref = mod.get_modref(minimal=True)
        return TopModRef(top=modref, sub=sub, tb=tbref)


def is_top(modcls: ModCls) -> bool:
    """Module is Direct Loadable."""
    if issubclass(modcls, AGenericTbMod):
        return modcls.build_dut.__qualname__ != AGenericTbMod.build_dut.__qualname__
    if issubclass(modcls, AConfigurableMod):
        if modcls.get_default_config is not AConfigurableMod.get_default_config:
            return True
        config_field = modcls.model_fields["config"]
        return config_field.default is not PydanticUndefined
    if issubclass(modcls, (AMod, ATbMod)):
        return True
    return modcls.build_top.__qualname__ != BaseMod.build_top.__qualname__


def get_tb(modcls: ModCls) -> TbType:
    """Module Testbench."""
    if issubclass(modcls, AGenericTbMod):
        return "Generic"
    if issubclass(modcls, ATbMod):
        return "Static"
    return ""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Array Type.
"""

from typing import Any, ClassVar

from .object import Light, PosArgs
from .slices import UP, Slice, SliceDirection
from .typebase import ACompositeType, BaseType


class ArrayType(ACompositeType, Light):
    """
    Array of `itemtype`.

    Args:
        itemtype (AType): Element Type.
        depth (int):    depth.

    Example:
        >>> import ucdp as u
        >>> mem = u.ArrayType(u.UintType(16), 10)
        >>> mem
        ArrayType(UintType(16), 10)

    Arrays can be nested and combined with all other types.

        >>> class BusType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add('data', u.UintType(8))
        ...         self._add('valid', u.BitType())
        ...         self._add('accept', u.BitType(), orientation=u.BWD)

        >>> structmatrix = u.ArrayType(u.ArrayType(BusType(), 10), 22)
        >>> structmatrix
        ArrayType(ArrayType(BusType(), 10), 22)

    Slicing:

        >>> structmatrix.slice_
        Slice('0:21')
        >>> structmatrix[0:15]
        ArrayType(ArrayType(BusType(), 10), 16)
        >>> structmatrix[3]
        ArrayType(BusType(), 10)
        >>> structmatrix[3][3]
        BusType()
    """

    itemtype: BaseType
    depth: Any
    left: Any = None
    right: Any = None
    direction: SliceDirection = UP
    packed: bool = False

    _posargs: ClassVar[PosArgs] = ("itemtype", "depth")

    def __init__(
        self,
        itemtype: BaseType,
        depth: Any,
        left=None,
        right=None,
        direction: SliceDirection = UP,
        packed: bool = False,
    ):
        if direction is UP:
            if (left == 0) and (right == depth - 1):
                left, right = None, None
        elif (right == 0) and (left == depth - 1):
            left, right = None, None
        super().__init__(itemtype=itemtype, depth=depth, left=left, right=right, direction=direction, packed=packed)  # type: ignore[call-arg]

    @property
    def slice_(self):
        """Get Slice of Matrix."""
        return Slice(left=self.left, right=self.right, width=self.depth, direction=self.direction)

    def is_connectable(self, other) -> bool:
        """
        Check For Valid Connection To `other`.

        Connections are only allowed to other :any:`ArrayType` of the same depth and type.

        >>> from ucdp import ArrayType, UintType, SintType
        >>> ArrayType(UintType(8), 8).is_connectable(ArrayType(UintType(8), 8))
        True
        >>> ArrayType(UintType(8), 8).is_connectable(ArrayType(UintType(8), 9))
        False
        >>> ArrayType(UintType(8), 8).is_connectable(ArrayType(SintType(8), 8))
        False
        """
        return (
            isinstance(other, ArrayType) and self.depth == other.depth and self.itemtype.is_connectable(other.itemtype)
        )

    def __getitem__(self, slice_):
        slice_ = Slice.cast(slice_)
        if slice_.width == 1:
            return self.itemtype
        return self.new(depth=slice_.width, left=slice_.left, right=slice_.right)

    @property
    def bits(self):
        """Size in Bits."""
        return self.depth * self.itemtype.bits



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Flip-Flop."""

from pytest import fixture

import ucdp as u


@fixture
def idents():
    """Some Ports."""
    return u.Idents(
        [
            u.Port(u.ClkRstAnType(), "main_i"),
            u.Port(u.UintType(8), "vec_a_i"),
            u.Signal(u.UintType(8), "vec_a_s"),
            u.Signal(u.UintType(4), "vec_b_s"),
            u.Signal(u.UintType(4), "vec_c_s"),
        ]
    )


def test_flipflop(idents):
    """Basics."""
    rst_an = u.Signal(u.RstAnType(), "rst_s")
    clk = u.Signal(u.ClkType(), "clk_s")
    flipflop = u.FlipFlop(targets=idents, sources=idents, rst_an=rst_an, clk=clk)

    assert flipflop.rst_an is rst_an
    assert flipflop.clk is clk
    assert flipflop.rst is None
    assert flipflop.ena is None

    # TESTME: rst_an type error
    # TESTME: clk type error


def test_flipflop_rst(idents):
    """Flip Flop with Reset."""

    # TESTME:


def test_flipflop_ena(idents):
    """Flip Flop with Ena."""

    # TESTME:



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Cache."""

import os
from unittest import mock

import ucdp as u
from ucdp.cache import Cache


def test_default(monkeypatch):
    """Default."""
    monkeypatch.delenv("UCDP_CACHE", raising=False)
    cache = Cache.init()
    assert cache.path
    assert cache.templates_path == cache.path / "templates"
    assert cache.path.exists()


def test_get_cache(tmp_path):
    """Get Cache."""
    with mock.patch.dict(os.environ, {"UCDP_CACHE": str(tmp_path)}):
        cache = Cache.init()

        test_get_cache.count = 0

        @cache.get_cache("foo")()
        def myfunc(a, b):
            test_get_cache.count += 1
            return a + b

        assert myfunc(1, 2) == 3
        assert test_get_cache.count == 1
        assert myfunc(1, 2) == 3
        assert test_get_cache.count == 1

        cache.clear()

        assert myfunc(1, 2) == 3
        assert test_get_cache.count == 2


def test_env(tmp_path):
    """Via Env Path."""
    with mock.patch.dict(os.environ, {"UCDP_CACHE": str(tmp_path)}):
        cache = Cache.init()
        assert cache.path == tmp_path
        assert cache.templates_path == tmp_path / "templates"


def test_env_invalid(tmp_path):
    """Via Invalid Env Path."""
    invalid_path = tmp_path / "invalid"
    invalid_path.touch()
    with mock.patch.dict(os.environ, {"UCDP_CACHE": str(invalid_path)}):
        cache = Cache.init()
        assert cache.path is None
        assert cache.templates_path is None


def test_env_disabled(tmp_path):
    """Caching Disabled."""
    with mock.patch.dict(os.environ, {"UCDP_CACHE": ""}):
        cache = Cache.init()
        assert cache.path is None
        assert cache.templates_path is None


def test_cache():
    """Test Cache is Exposed."""
    assert isinstance(u.CACHE, Cache)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Test Assigns."""

from pytest import fixture

import ucdp as u


def _check_assigns(assigns, pairs):
    """Check Assignments."""
    assert tuple((assign.target, assign.source) for assign in assigns) == pairs


def _check(idents, dest, src, error=None):
    assigns = u.Assigns(targets=idents)
    if not error:
        assigns.set(dest, src)
        if dest.direction != u.IN:
            _check_assigns(assigns, ((dest, src),))
        else:
            _check_assigns(assigns, ((src, dest),))


class MstType(u.AStructType):
    """Mst."""

    def _build(self):
        self._add("trans", u.UintType(2))
        self._add("wdata", u.UintType(8))
        self._add("rdata", u.UintType(8), u.BWD)
        self._add("ready", u.BitType(), u.BWD)

    @staticmethod
    def cast(other):
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, SlvType):
            # Drive a Mst input with Slv signals
            yield "", ""
            yield "trans", "trans"
            yield "wdata", "wdata"
            yield "rdata", "rdata"
            yield "ready", "readyout"
        return NotImplemented


class SlvType(u.AStructType):
    """Slave."""

    def _build(self):
        self._add("trans", u.UintType(2))
        self._add("wdata", u.UintType(8))
        self._add("rdata", u.UintType(8), u.BWD)
        self._add("readyout", u.BitType(), u.BWD)
        self._add("ready", u.BitType())
        self._add("sel", u.BitType())

    @staticmethod
    def cast(other):
        """
        How to cast an input of type `self` from a value of type `other`.

        `self = cast(other)`
        """
        if isinstance(other, MstType):
            # Drive a Slv input with Mst signals
            yield "", ""
            yield "trans", "trans"
            yield "wdata", "wdata"
            yield "rdata", "rdata"  # BWD
            yield "readyout", "ready"  # BWD
            yield "ready", u.const("1b1")
            yield "sel", "ternary(trans > const('1b0'), const('1b1'), const('1b0'))"
        return NotImplemented


@fixture
def idents():
    """Idents For Testing."""
    return u.Idents(
        [
            # vectors
            u.Port(u.UintType(8), "uint_src_i"),
            u.Port(u.UintType(8), "uint_src_o"),
            u.Port(u.UintType(8), "uint_src_io"),
            u.Signal(u.UintType(8), "uint_src_s"),
            u.Port(u.UintType(8), "uint_dest_i"),
            u.Port(u.UintType(8), "uint_dest_o"),
            u.Port(u.UintType(8), "uint_dest_io"),
            u.Signal(u.UintType(8), "uint_dest_s"),
            # master
            u.Port(MstType(), "mst_src_i"),
            u.Port(MstType(), "mst_src_o"),
            u.Port(MstType(), "mst_src_io"),
            u.Signal(MstType(), "mst_src_s"),
            u.Port(MstType(), "mst_dest_i"),
            u.Port(MstType(), "mst_dest_o"),
            u.Port(MstType(), "mst_dest_io"),
            u.Signal(MstType(), "mst_dest_s"),
            # slave
            u.Port(SlvType(), "slv_src_i"),
            u.Port(SlvType(), "slv_src_o"),
            u.Port(SlvType(), "slv_src_io"),
            u.Signal(SlvType(), "slv_src_s"),
            u.Port(SlvType(), "slv_dest_i"),
            u.Port(SlvType(), "slv_dest_o"),
            u.Port(SlvType(), "slv_dest_io"),
            u.Signal(SlvType(), "slv_dest_s"),
        ]
    )


@fixture
def intidents():
    """Idents For Testing."""
    return u.Idents(
        [
            # vectors
            u.Port(u.UintType(8), "uint_src_i"),
            u.Port(u.UintType(8), "uint_src_o"),
            u.Port(u.UintType(8), "uint_src_io"),
            u.Signal(u.UintType(8), "uint_src_s"),
            u.Port(u.UintType(8), "uint_dest_i"),
            u.Port(u.UintType(8), "uint_dest_o"),
            u.Port(u.UintType(8), "uint_dest_io"),
            u.Signal(u.UintType(8), "uint_dest_s"),
            u.Port(u.SintType(8), "sint_src_i"),
            u.Port(u.SintType(8), "sint_src_o"),
            u.Port(u.SintType(8), "sint_src_io"),
            u.Signal(u.SintType(8), "sint_src_s"),
            u.Port(u.SintType(8), "sint_dest_i"),
            u.Port(u.SintType(8), "sint_dest_o"),
            u.Port(u.SintType(8), "sint_dest_io"),
            u.Signal(u.SintType(8), "sint_dest_s"),
        ]
    )


def test_fwd(idents):
    """Forward."""
    expr = u.const("8'hfc")

    # src=signal
    _check(idents, idents["uint_dest_s"], idents["uint_src_s"])
    _check(idents, idents["uint_dest_i"], idents["uint_src_s"])
    _check(idents, idents["uint_dest_o"], idents["uint_src_s"])
    _check(idents, idents["uint_dest_io"], idents["uint_src_s"])
    _check(idents, expr, idents["uint_src_s"], error="expression")
    # src=IN
    _check(idents, idents["uint_dest_s"], idents["uint_src_i"])
    _check(idents, idents["uint_dest_i"], idents["uint_src_i"], error="direction")
    _check(idents, idents["uint_dest_o"], idents["uint_src_i"])
    _check(idents, idents["uint_dest_io"], idents["uint_src_i"])
    _check(idents, expr, idents["uint_src_i"], error="expression")
    # src=OUT
    _check(idents, idents["uint_dest_s"], idents["uint_src_o"])
    _check(idents, idents["uint_dest_i"], idents["uint_src_o"])
    _check(idents, idents["uint_dest_o"], idents["uint_src_o"])
    _check(idents, idents["uint_dest_io"], idents["uint_src_o"])
    _check(idents, expr, idents["uint_src_o"], error="expression")
    # src=INOUT
    _check(idents, idents["uint_dest_s"], idents["uint_src_io"])
    _check(idents, idents["uint_dest_i"], idents["uint_src_io"])
    _check(idents, idents["uint_dest_o"], idents["uint_src_io"])
    _check(idents, idents["uint_dest_io"], idents["uint_src_io"])
    _check(idents, expr, idents["uint_src_io"], error="expression")
    # src=expr
    _check(idents, idents["uint_dest_s"], expr)
    _check(idents, idents["uint_dest_i"], expr, error="expression")
    _check(idents, idents["uint_dest_o"], expr)
    _check(idents, idents["uint_dest_io"], expr)
    _check(idents, expr, expr, error="expression")


def test_modassign_slv2mst_cast(idents):
    """Test Bi-Directional Assignments within module (assign statements) and casting."""
    assigns = u.Assigns(targets=idents)
    # Mst = cast(Slv)
    assigns.set(idents["mst_dest_o"], idents["slv_src_i"], cast=True)
    _check_assigns(
        assigns,
        (
            (idents["mst_dest_o"], idents["slv_src_i"]),
            (idents["mst_dest_trans_o"], idents["slv_src_trans_i"]),
            (idents["mst_dest_wdata_o"], idents["slv_src_wdata_i"]),
            (idents["slv_src_rdata_o"], idents["mst_dest_rdata_i"]),
            (idents["slv_src_readyout_o"], idents["mst_dest_ready_i"]),
        ),
    )


def test_modassign_mst2slv_cast(idents):
    """Test Bi-Directional Assignments within module (assign statements) and casting."""
    assigns = u.Assigns(targets=idents)
    # Mst = cast(Slv)
    assigns.set(idents["slv_dest_o"], idents["mst_src_i"], cast=True)
    _check_assigns(
        assigns,
        (
            (idents["mst_src_rdata_o"], idents["slv_dest_rdata_i"]),
            (idents["mst_src_ready_o"], idents["slv_dest_readyout_i"]),
            (idents["slv_dest_o"], idents["mst_src_i"]),
            (idents["slv_dest_trans_o"], idents["mst_src_trans_i"]),
            (idents["slv_dest_wdata_o"], idents["mst_src_wdata_i"]),
            (idents["slv_dest_ready_o"], u.ConstExpr(u.BitType(default=1))),
            (
                idents["slv_dest_sel_o"],
                u.TernaryExpr(
                    u.BoolOp(u.Port(u.UintType(2), "mst_src_trans_i", direction=u.IN), ">", u.ConstExpr(u.BitType())),
                    u.ConstExpr(u.BitType(default=1)),
                    u.ConstExpr(u.BitType()),
                ),
            ),
        ),
    )



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Router."""

import re
from collections.abc import Iterable

from .expr import Expr
from .iterutil import split
from .note import Note
from .object import LightObject, model_validator
from .signal import BaseSignal

_RE_PATH = re.compile(
    r"(?P<beg>(((cast)|(optcast)|(create))\()*)"
    r"(?P<path>((([a-zA-Z0-9_]+)|(\.\.))/)*)?(?P<expr>[^/]+?)"
    r"(?P<end>\)*)"
)


class RoutePath(LightObject):
    """
    Routing Path.

    Attributes:
        expr: Expression
        path: Module Path
        create: Create signal/port referenced by `expr`
        cast: `True` (required), `None` (optional) or `False` (forbidden) type casting.
    """

    expr: str | BaseSignal | Note
    path: str | None = None
    create: bool = False
    cast: bool | None = False

    @model_validator(mode="after")
    def __post_init(self) -> "RoutePath":
        if not (not self.create or not self.cast):
            raise ValueError(f"{self}: [opt]cast() and create() are mutually exclusive.")
        return self

    @property
    def parts(self) -> tuple[str, ...]:
        """Path Parts."""
        if not self.path:
            return ()
        return tuple(self.path.split("/"))

    def __str__(self):
        if self.path:
            res = f"{self.path}/{self.expr}"
        else:
            res = str(self.expr)
        if self.cast is True:
            res = f"cast({res})"
        elif self.cast is None:
            res = f"optcast({res})"
        if self.create:
            res = f"create({res})"
        return res


Routeable = RoutePath | Expr | str | Note
Routeables = Routeable | Iterable[Routeable]


def parse_routepath(value: Routeable, basepath: str | None = None) -> RoutePath:
    """
    Parse Path.

    >>> parse_routepath('clk_i')
    RoutePath(expr='clk_i')
    >>> parse_routepath('u_sub/u_bar/clk_i')
    RoutePath(expr='clk_i', path='u_sub/u_bar')
    >>> parse_routepath('create(port_i)')
    RoutePath(expr='port_i', create=True)
    >>> parse_routepath('cast(port_i)')
    RoutePath(expr='port_i', cast=True)
    >>> parse_routepath('optcast(port_i)')
    RoutePath(expr='port_i', cast=None)
    >>> parse_routepath('u_sub/u_bar/port_i')
    RoutePath(expr='port_i', path='u_sub/u_bar')
    >>> parse_routepath('../port_i')
    RoutePath(expr='port_i', path='..')

    >>> parse_routepath('')
    Traceback (most recent call last):
    ...
    ValueError: Invalid route path ''
    >>> parse_routepath('/')
    Traceback (most recent call last):
      ...
    ValueError: Invalid route path '/'
    >>> parse_routepath('u_sub/u_bar/')
    Traceback (most recent call last):
      ...
    ValueError: Invalid route path 'u_sub/u_bar/'
    >>> parse_routepath('.../')
    Traceback (most recent call last):
      ..
    ValueError: Invalid route path '.../'
    """
    if isinstance(value, RoutePath):
        return value
    if isinstance(value, (Expr, Note)):
        return RoutePath(expr=value, path=basepath or None)
    mat = _RE_PATH.fullmatch(value)
    if mat:
        path = mat.group("path")
        if path:
            path = path[:-1]
        if path and basepath:
            path = f"{basepath}/{path}"
        else:
            path = basepath or path
        path = path or None
        expr = mat.group("expr")
        # create = bool(mat.group("crt"))
        beg = (mat.group("beg") or "").split("(")
        end = (mat.group("end") or "").split(")")
        if len(beg) == len(end):
            create = "create" in beg
            if "optcast" in beg:
                cast = None
            else:
                cast = "cast" in beg
            return RoutePath(expr=expr, path=path, create=create, cast=cast)
    raise ValueError(f"Invalid route path {value!r}")


def parse_routepaths(routepaths: Routeables | None, basepath: str | None = None) -> tuple[RoutePath, ...]:
    """
    Parse `routepaths`.

    >>> parse_routepaths('clk_i')
    (RoutePath(expr='clk_i'),)
    >>> parse_routepaths((RoutePath(expr='clk_i'),))
    (RoutePath(expr='clk_i'),)
    >>> parse_routepaths(RoutePath(expr='clk_i'))
    (RoutePath(expr='clk_i'),)
    >>> parse_routepaths('clk_i; rst_an_i')
    (RoutePath(expr='clk_i'), RoutePath(expr='rst_an_i'))
    >>> parse_routepaths('clk_i, rst_an_i')
    (RoutePath(expr='clk_i, rst_an_i'),)
    """
    if not routepaths:
        return ()
    if isinstance(routepaths, (RoutePath, Expr, Note)):
        return (parse_routepath(routepaths, basepath=basepath),)
    if isinstance(routepaths, str):
        return tuple(parse_routepath(path, basepath=basepath) for path in split(routepaths))
    return tuple(parse_routepath(path, basepath=basepath) for path in routepaths)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


"""
Driver Tracking.

"""

from collections.abc import Iterator
from typing import Annotated

from pydantic import BeforeValidator

from .exceptions import MultipleDriverError
from .expr import Expr, SliceOp
from .logging import LOGGER
from .note import Note
from .object import Field, Object
from .signal import BaseSignal
from .typebase import BaseScalarType

Target = BaseSignal | SliceOp
"""Assignment Target."""

Source = Expr | Note
"""Assignment Source."""


def check_target(target: Target) -> None:
    """
    Check `target`.

    Only ports, signals of any type OR slicing of native types are supported.
    """
    if isinstance(target, BaseSignal):
        return

    if isinstance(target, SliceOp):
        if isinstance(target.one.type_, BaseScalarType):
            return
        raise ValueError(f"Invalid target {target!r}. Slicing is only supported on native types")


ToTarget = Annotated[
    Target,
    BeforeValidator(lambda x: check_target(x)),
]


class Drivers(Object):
    """
    Drivers.

    This container tracks multiple drivers as every added signal is only allowed to be driven once.

    Attributes:
        drivers: Dictionary with drivers.
    """

    drivers: dict[str, Source] = Field(default_factory=dict, init=False)

    def set(self, target: BaseSignal, source: Source, overwrite: bool = False):
        """Set Driver for Target."""
        drivers = self.drivers
        name = target.name
        LOGGER.debug("driver target=%s source=%s", target, source)
        if not overwrite and name in drivers:
            raise MultipleDriverError(f"'{source}' already driven by '{drivers[name]}'")
        drivers[name] = source

    def __iter__(self) -> Iterator[tuple[str, Source]]:  # type: ignore[override]
        yield from self.drivers.items()



# MIT License
#
# Copyright (c) 2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Note.

This note can be used instead of an expression.
"""

from .object import LightObject


class Note(LightObject):
    """
    Note on Assignments Instead Of Value.

    Attributes:
        note: Note.

    ??? Example "Note Examples"
        Example.

            >>> import ucdp as u
            >>> u.OPEN
            Note(note='OPEN')
    """

    note: str

    def __str__(self):
        return self.note


class Default(Note):
    """
    Note on Assignments With Default Value on Inputs.

    Attributes:
        note: Note.

    ??? Example "Note Examples"
        Example.

            >>> import ucdp as u
            >>> u.TODO
            Default(note='TODO')
            >>> u.DEFAULT
            Default(note='DEFAULT')
            >>> u.UNUSED
            Default(note='UNUSED')
    """


def note(note: str) -> Note:
    """Create Note."""
    return Note(note=note)


OPEN = Note(note="OPEN")
"""Open Note."""

TODO = Default(note="TODO")
"""Todo Note - Use Defaults."""

DEFAULT = Default(note="DEFAULT")
"""Use Defaults."""

UNUSED = Default(note="UNUSED")
"""UNUSED Note  - Use Defaults."""



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Command Line Interface."""

import logging
import sys
from collections import defaultdict
from collections.abc import Iterable
from logging import StreamHandler
from pathlib import Path
from typing import Literal

import click
from click_bash42_completion import patch
from pydantic import BaseModel, ConfigDict
from rich import box
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import pprint
from rich.table import Table

from . import consts
from ._cligroup import MainGroup
from ._logging import HasErrorHandler
from .cache import CACHE
from .cliutil import (
    Exit,
    PathType,
    arg_template_filepaths,
    arg_top,
    arg_tops,
    auto_path,
    defines2data,
    opt_check,
    opt_clean,
    opt_create,
    opt_defines,
    opt_dry_run,
    opt_file,
    opt_filelist,
    opt_filepath,
    opt_local,
    opt_maxlevel,
    opt_maxworkers,
    opt_path,
    opt_show_diff,
    opt_tag,
    opt_target,
    opt_topsfile,
    read_file,
)
from .create import TB_MAP, TYPE_CHOICES, CreateInfo
from .create import create as create_
from .fileset import FileSet
from .finder import find
from .generate import Generator, clean, get_makolator, render_generate, render_inplace
from .iterutil import namefilter
from .loader import load
from .modfilelist import iter_modfilelists
from .modtopref import PAT_TOPMODREF, TopModRef
from .pathutil import absolute, relative
from .top import Top
from .util import LOGGER, guess_path

patch()


_LOGLEVELMAP = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


class Ctx(BaseModel):
    """Command Line Context."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    console: Console
    has_error_handler: HasErrorHandler | None = None

    verbose: int = 0
    no_cache: bool = False
    no_color: bool | None = None

    @staticmethod
    def create(no_color: bool | None = None, **kwargs) -> "Ctx":
        """Create."""
        console = Console(log_time=False, log_path=False, no_color=no_color)
        has_error_handler = HasErrorHandler()
        return Ctx(console=console, has_error_handler=has_error_handler, no_color=no_color, **kwargs)

    def __enter__(self):
        # Logging
        level = _LOGLEVELMAP.get(self.verbose, logging.DEBUG)
        if not self.no_color:
            handler = RichHandler(
                show_time=False,
                show_path=False,
                rich_tracebacks=True,
                console=Console(stderr=True, no_color=self.no_color),
            )
            format_ = "%(message)s"
        else:
            handler = StreamHandler(stream=sys.stderr)
            format_ = "%(levelname)s %(message)s"
        handlers = [handler, self.has_error_handler]
        logging.basicConfig(level=level, format=format_, handlers=handlers)

        # Cache
        if self.no_cache:
            CACHE.disable()

        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass


@click.group(cls=MainGroup, context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-v", "--verbose", count=True, help="Increase Verbosity.")
@click.option("-C", "--no-cache", is_flag=True, help="Disable Caching.")
@click.option("--no-color", is_flag=True, help="Disable Coloring.", envvar="UCDP_NO_COLOR")
@click.version_option()
@click.pass_context
def ucdp(ctx, verbose=0, no_cache=False, no_color=False):
    """Unified Chip Design Platform."""
    ctx.obj = ctx.with_resource(Ctx.create(verbose=verbose, no_cache=no_cache, no_color=no_color))


pass_ctx = click.make_pass_decorator(Ctx)


@ucdp.result_callback()
@pass_ctx
def _end(ctx: Ctx, *args, **kwargs):
    if ctx.has_error_handler.has_errors:
        raise Exit(ctx.console, "FAILED")


def get_group(help=None):  # pragma: no cover
    """Create Command Group."""

    @click.group(help=help)
    @click.pass_context
    def group(ctx):
        ctx.obj = Ctx(console=Console(log_time=False, log_path=False))

    return group


def load_top(ctx: Ctx, top: str | TopModRef, paths: Iterable[str | Path], quiet: bool = False) -> Top:
    """Load Top Module."""
    lpaths = [Path(path) for path in paths]
    # Check if top seems to be some kind of file path
    topmodref = TopModRef.cast(guess_path(top) or top) if isinstance(top, str) else top
    if quiet:
        return load(topmodref, paths=lpaths)
    with ctx.console.status(f"Loading '{topmodref!s}'"):
        result = load(topmodref, paths=lpaths)
    ctx.console.log(f"'{topmodref!s}' checked.")
    return result


@ucdp.command(
    help=f"""
Load Data Model and Check.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_top
@opt_path
@click.option("--stat", default=False, is_flag=True, help="Show Statistics.")
@pass_ctx
def check(ctx, top, path, stat=False):
    """Check."""
    top = load_top(ctx, top, path)
    if stat:
        print("Statistics:")
        for name, value in top.get_stat().items():
            print(f"  {name}: {value}")


@ucdp.command(
    help=f"""
Load Data Model and Generate Files.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_tops
@opt_path
@opt_filelist
@opt_target
@opt_show_diff
@opt_maxworkers
@opt_defines
@opt_local
@opt_topsfile
@opt_check
@opt_create
@opt_clean
@pass_ctx
def gen(
    ctx,
    tops,
    path,
    filelist,
    target=None,
    show_diff=False,
    maxworkers=None,
    define=None,
    local=None,
    check=False,
    create=False,
    clean=False,
    tops_file=None,
):
    """Generate."""
    tops = list(tops)
    for filepath in tops_file or []:
        tops.extend(read_file(filepath))
    makolator = get_makolator(show_diff=show_diff, paths=path, create=create)
    data = defines2data(define)
    filelist = filelist or ["*"]
    with Generator(makolator=makolator, maxworkers=maxworkers, check=check) as generator:
        for info in find(path, patterns=tuple(tops), local=local, is_top=True):
            try:
                top = load_top(ctx, info.topmodref, path)
            except Exception as exc:
                LOGGER.warning(f"Cannot load '{info.topmodref}'")
                LOGGER.warning(str(exc))
                LOGGER.warning(f"Debug with '{consts.CLINAME} check {info.topmodref}'")
                continue
            for item in filelist:
                generator.generate(top, item, target=target, data=data, clean=clean)


@ucdp.command(
    help=f"""
Load Data Model and Render Template and Create File.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'

TEMPLATE_FILEPATHS: Templates to render. Environment Variable 'UCDP_TEMPLATE_FILEPATHS'
                    Templates in `templates` folders are found automatically.

GENFILE: Generated File.
"""
)
@arg_top
@opt_path
@arg_template_filepaths
@click.argument("genfile", type=PathType, shell_complete=auto_path, nargs=1)
@opt_show_diff
@opt_defines
@opt_create
@pass_ctx
def rendergen(ctx, top, path, template_filepaths, genfile, show_diff=False, define=None, create=False):
    """Render Generate."""
    top = load_top(ctx, top, path)
    makolator = get_makolator(show_diff=show_diff, paths=path, create=create)
    data = defines2data(define)
    render_generate(top, template_filepaths, genfile=genfile, makolator=makolator, data=data)


@ucdp.command(
    help=f"""
Load Data Model and Render Template and Update File.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'

TEMPLATE_FILEPATHS: Templates to render. Environment Variable 'UCDP_TEMPLATE_FILEPATHS'
                    Templates in `templates` folders are found automatically.

INPLACEFILE: Inplace File.
"""
)
@arg_top
@opt_path
@arg_template_filepaths
@click.argument("inplacefile", type=PathType, shell_complete=auto_path, nargs=1)
@opt_show_diff
@opt_defines
@opt_create
@click.option("--ignore_unknown", "-i", default=False, is_flag=True, help="Ignore Unknown Placeholder.")
@pass_ctx
def renderinplace(
    ctx, top, path, template_filepaths, inplacefile, show_diff=False, define=None, ignore_unknown=False, create=False
):
    """Render Inplace."""
    top = load_top(ctx, top, path)
    makolator = get_makolator(show_diff=show_diff, paths=path, create=create)
    data = defines2data(define)
    render_inplace(
        top,
        template_filepaths,
        inplacefile=inplacefile,
        makolator=makolator,
        data=data,
        ignore_unknown=ignore_unknown,
    )


@ucdp.command(
    help=f"""
Load Data Model and REMOVE Generated Files.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_top
@opt_path
@opt_filelist
@opt_target
@opt_show_diff
@opt_dry_run
@opt_maxworkers
@pass_ctx
def cleangen(ctx, top, path, filelist, target=None, show_diff=False, maxworkers=None, dry_run=False):
    """Clean Generated Files."""
    top = load_top(ctx, top, path)
    makolator = get_makolator(show_diff=show_diff, paths=path)
    for item in filelist or ["*"]:
        clean(top, item, target=target, makolator=makolator, maxworkers=maxworkers, dry_run=dry_run)


@ucdp.command(
    help=f"""
Load Data Model and Generate File List.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_top
@opt_path
@opt_filelist
@opt_target
@opt_file
@pass_ctx
def filelist(ctx, top, path, filelist, target=None, file=None):
    """File List."""
    # Load quiet, otherwise stdout is messed-up
    top = load_top(ctx, top, path, quiet=True)
    for item in filelist or ["*"]:
        fileset = FileSet.from_mod(top.mod, item, target=target)
        for line in fileset:
            print(line, file=file)


@ucdp.command(
    help=f"""
Load Data Model and Show File Information

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_top
@opt_path
@opt_filelist
@opt_target
@opt_maxlevel
@click.option("--minimal", "-m", default=False, is_flag=True, help="Skip defaults.")
@opt_file
@pass_ctx
def fileinfo(ctx, top, path, filelist, target=None, maxlevel=None, minimal=False, file=None):
    """File List."""
    # Load quiet, otherwise stdout is messed-up
    top = load_top(ctx, top, path, quiet=True)
    console = Console(file=file) if file else ctx.console
    for item in filelist or ["*"]:
        data = defaultdict(list)
        for mod, modfilelist in iter_modfilelists(top.mod, item, target=target, maxlevel=maxlevel):
            data[str(mod)].append(modfilelist.model_dump(exclude_defaults=minimal))
        pprint(dict(data), indent_guides=False, console=console)


@ucdp.command(
    help=f"""
              List Available Data Models.

              PATTERN: Limit list to these modules only.

              Examples:

                {consts.CLINAME} ls

                {consts.CLINAME} ls -n

                {consts.CLINAME} ls glbl_lib*
              """
)
@arg_tops
@opt_path
@click.option("--names", "-n", default=False, is_flag=True, help="Just print names")
@click.option("--top/--no-top", "-t/-T", default=None, is_flag=True, help="List loadable top modules only.")
@click.option("--tb/--no-tb", "-b/-B", default=None, is_flag=True, help="List testbench modules only.")
@click.option("--generic-tb", "-g", default=False, is_flag=True, help="List Generic Testbench modules only.")
@opt_local
@click.option("--base", "-A", default=False, is_flag=True, help="Show Base Classes.")
@click.option("--filepath", "-f", default=False, is_flag=True, help="Show File Path.")
@click.option("--abs-filepath", "-F", default=False, is_flag=True, help="Show Absolute File Path.")
@opt_tag
@pass_ctx
def ls(  # noqa: C901
    ctx,
    path=None,
    tops=None,
    names=False,
    top=None,
    tb=None,
    local=None,
    generic_tb=False,
    tag=None,
    base=False,
    filepath=False,
    abs_filepath=False,
):
    """List Modules."""
    with ctx.console.status("Searching"):
        infos = find(path, patterns=tops or ["*"], local=local)
    if top is not None:
        infos = [info for info in infos if info.is_top == top]
    if tb is not None:
        infos = [info for info in infos if bool(info.tb) == tb]
    if generic_tb:
        infos = [info for info in infos if info.tb == "Generic"]
    if tag:
        filter_ = namefilter(tag)
        infos = [info for info in infos if any(filter_(tag) for tag in info.tags)]

    def fill_row(row, info):
        if base:
            row.append(info.modbasecls.__name__)
        if filepath:
            row.append(str(relative(info.filepath)))
        if abs_filepath:
            row.append(str(info.filepath))

    if names:
        for info in infos:
            row = [info.topmodref]
            fill_row(row, info)
            print(*row)
    else:
        table = Table(expand=filepath or abs_filepath, box=box.MARKDOWN)
        table.add_column("Reference")
        table.add_column("Top", justify="center")
        table.add_column("Tb ", justify="center")
        table.add_column("Tags")
        if base:
            table.add_column("Bases on", justify="right")
        if filepath:
            table.add_column("Filepath")
        if abs_filepath:
            table.add_column("Absolute Filepath")
        for info in infos:
            row = [
                str(info.topmodref),
                "X" if info.is_top else "",
                "X" if info.tb else "",
                ",".join(sorted(info.tags)),
            ]
            fill_row(row, info)
            table.add_row(*row)
        ctx.console.print(table)


@ucdp.command(
    help=f"""
Load Data Model and Module Information.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_tops
@opt_path
@opt_local
@click.option("--top", "-t", default=None, is_flag=True, help="List loadable top modules only.")
@click.option("--sub", "-S", default=False, is_flag=True, help="Show Submodules.")
@pass_ctx
def modinfo(ctx, tops, path, local, top, sub):
    """Module Information."""
    sep = ""
    for info in find(path, patterns=tops, local=local, is_top=top):
        try:
            top = load_top(ctx, info.topmodref, path, quiet=True)
        except Exception as exc:
            LOGGER.warning(str(exc))
            continue
        print(sep + top.mod.get_info(sub=sub))
        sep = "\n\n"


@ucdp.command(
    help=f"""
Load Data Model and Show Module Overview.

TOP: Top Module. {PAT_TOPMODREF}. Environment Variable 'UCDP_TOP'
"""
)
@arg_top
@opt_path
@click.option("--minimal", "-m", default=False, is_flag=True, help="Skip modules without specific details")
@opt_filepath
@opt_tag
@pass_ctx
def overview(ctx, top, path, minimal=False, file=None, tag=None):
    """Overview."""
    # Load quiet, otherwise stdout is messed-up
    top = load_top(ctx, top, path, quiet=True)
    data = {"minimal": minimal, "tags": tag}
    render_generate(top, [consts.PATH / "ucdp-templates" / "overview.txt.mako"], genfile=file, data=data, no_stat=True)


@ucdp.group(context_settings={"help_option_names": ["-h", "--help"]})
def info():
    """Information."""


@info.command()
@pass_ctx
def examples(ctx):
    """Path to Examples."""
    examples_path = Path(__file__).parent / "examples"
    print(str(examples_path))


@info.command()
@opt_path
@pass_ctx
def template_paths(ctx, path):
    """Template Paths."""
    makolator = get_makolator(paths=path)
    for template_path in makolator.config.template_paths:
        print(str(template_path))


@ucdp.command(
    help="""
Create Datamodel Skeleton.
"""
)
@click.option("--module", "-m", prompt=True, help="Name of the Module")
@click.option("--library", "-l", default=absolute(Path()).name, prompt=True, help="Name of the Library")
@click.option("--regf/--no-regf", "-r/-R", default=True, help="Make use of a Register File")
@click.option("--descr", "-d", default="", help="Description")
@click.option("--flavour", "-F", type=click.Choice(TYPE_CHOICES, case_sensitive=False), help="Choose a Module Flavour")
@click.option("--tb/--no-tb", "-t/-T", default=None, help="Create testbench for design module")
@click.option("--force", "-f", is_flag=True, help="Overwrite existing files")
@pass_ctx
def create(
    ctx,
    module,
    library,
    regf,
    descr,
    flavour,
    tb,
    force,
):
    """Let The User Type In The Name And Library Of The File."""
    if flavour is None:
        flavour = prompt_flavour()
        ctx.console.print(f"\nYou choose the flavour [bold blue]{flavour}[/bold blue].\n")

    info = CreateInfo(module=module, library=library, regf=regf, descr=descr, flavour=flavour)

    if info.is_tb:
        if not module.endswith("_tb"):
            LOGGER.warning(f"Your testbench module name {module!r} does not end with '_tb'")

    else:
        if module.endswith("_tb"):
            LOGGER.warning(f"Your design module name {module!r} ends with '_tb'")
        if tb is None:
            answer = click.prompt(
                "Do you want to create a corresponding testbench? (y)es. (n)o.",
                type=click.Choice(["y", "n"]),
                default="y",
            )
            tb = answer == "y"
            ctx.console.print("")
        if tb:
            tbinfo = CreateInfo(module=f"{module}_tb", library=library, regf=regf, descr=descr, flavour=TB_MAP[flavour])
            create_(tbinfo, force)
    create_(info, force)


Type = Literal["AConfigurableMod", "AConfigurableTbMod", "AGenericTbMod", "AMod", "ATailoredMod", "ATbMod"]


def prompt_flavour() -> Type:
    """Let The User Choose The Type Of The File."""
    answer = click.prompt("Do you want to build a (d)esign or (t)estbench?", type=click.Choice(["d", "t"]), default="d")
    if answer == "d":
        answer = click.prompt(
            "Does your design vary more than what `parameter` can cover? (y)es. (n)o.", type=click.Choice(["y", "n"])
        )
        if answer == "y":
            answer = click.prompt(
                "Do you want to use a (c)onfig or (t) shall the parent module tailor the functionality?",
                type=click.Choice(["c", "t"]),
            )
            if answer == "c":
                flavour_ = "AConfigurableMod"

            else:
                flavour_ = "ATailoredMod"

        else:
            flavour_ = "AMod"

    else:
        answer = click.prompt(
            "Do you want to build a generic testbench which tests similar modules? (y)es. (n)o.",
            type=click.Choice(["y", "n"]),
        )
        if answer == "y":
            answer = click.prompt(
                "Do you want to automatically adapt your testbench to your (g) dut or use a (c)onfig?",
                type=click.Choice(["g", "c"]),
            )
            if answer == "c":
                flavour_ = "AConfigurableTbMod"

            else:
                flavour_ = "AGenericTbMod"

        else:
            flavour_ = "ATbMod"

    return flavour_



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""File List Parser."""

import re
from collections.abc import Iterable
from pathlib import Path

from .logging import LOGGER
from .object import Object
from .pathutil import improved_glob, improved_resolve

_RE_COMMENT = re.compile(r"\A(.*?)(\s*(#|//).*)\Z")
_RE_FILELIST = re.compile(r"\A-([fF])\s+(.*?)\Z")
_RE_INCDIR = re.compile(r'\A[+\-]incdir[+\-\s]"?(?P<incdir>.*?)"?\Z')
_RE_FILE = re.compile(r"\A((-sv|-v)\s+)?(?P<filepath>[^+-].*?)\Z")


class FileListParser(Object):
    """File List Parser."""

    def parse_file(  # type: ignore[override]
        self,
        filepaths: list[Path],
        inc_dirs: list[Path],
        filepath: Path,
        replace_envvars: bool = False,
        glob: bool = False,
    ):
        """Read File List File.

        Args:
            filepaths: File Paths Container.
            inc_dirs: Include Directories Container.
            filepath: File to be parsed.
            replace_envvars: Resolve Environment Variables.
            glob: Expand wildcards.
        """
        with filepath.open(encoding="utf-8") as file:
            basepath = filepath.parent
            self.parse(filepaths, inc_dirs, basepath, file, glob=glob)

        with filepath.open(encoding="utf-8") as file:
            basepath = filepath.parent
            self.parse(
                filepaths,
                inc_dirs,
                basepath,
                file,
                replace_envvars=replace_envvars,
                context=str(filepath),
                glob=glob,
            )

    def parse(  # noqa: C901
        self,
        filepaths: list[Path],
        inc_dirs: list[Path],
        basedir: Path,
        items: Iterable[str | Path],
        replace_envvars: bool = False,
        context: str = "",
        glob: bool = False,
    ):
        """File List File.

        Args:
            filepaths: File Paths Container.
            inc_dirs: Include Directories Container.
            basedir: Base Directory for Relative Paths.
            items: Items to be parsed.
            replace_envvars: Resolve Environment Variables.
            context: Context for error reporting.
            glob: Expand wildcards.
        """
        for lineno, item in enumerate(items, 1):
            line = str(item).strip()
            # comment
            mat = _RE_COMMENT.match(line)
            if mat:
                line = mat.group(1).strip()
            if not line:
                continue
            # -f
            mat = _RE_FILELIST.match(line)
            if mat:
                filelistpath = self.resolve(basedir, Path(mat.group(2)))
                self.parse_file(filepaths, inc_dirs, filelistpath, glob=glob)
                continue
            # -incdir
            mat = _RE_INCDIR.match(line)
            if mat:
                incdir = self.normalize(basedir, Path(mat.group("incdir")), replace_envvars)
                if incdir not in inc_dirs:
                    inc_dirs.append(incdir)
                continue
            # file
            mat = _RE_FILE.match(line)
            if mat:
                if glob:
                    items = improved_glob(Path(mat.group("filepath")), basedir=basedir)
                else:
                    items = [Path(mat.group("filepath"))]
                for sitem in items:
                    filepath = self.normalize(basedir, sitem, replace_envvars)
                    if filepath not in filepaths:
                        filepaths.append(filepath)
                continue
            LOGGER.warning("%s:%d Cannot parse %s", context, lineno, line)

    def resolve(self, basedir: Path, path: Path) -> Path:
        """Return Valid Filepath With Resolved Environment Variables.

        Args:
            basedir: Base Directory.
            path: Path.
        """
        return improved_resolve(path, basedir=basedir, replace_envvars=True, strict=True)

    def normalize(self, basedir: Path, path: Path, replace_envvars: bool) -> Path:
        """Return Normalized Filepaths for File List.

        Args:
            basedir: Base Directory.
            path: Path.
            replace_envvars: Resolve Environment Variables.
        """
        return improved_resolve(path, basedir=basedir, replace_envvars=replace_envvars)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Module Constant.

??? Example "Constant Examples"
    Usage:

        >>> from tabulate import tabulate
        >>> import ucdp as u
        >>> u.Const(u.UintType(6), "const_p")
        Const(UintType(6), 'const_p')

    Also complex types can simply be used:

        >>> class AType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("req", u.BitType())
        ...         self._add("data", u.ArrayType(u.UintType(16), 5))
        ...         self._add("ack", u.BitType(), u.BWD)
        >>> class MType(u.AEnumType):
        ...     keytype: u.AScalarType = u.UintType(2)
        ...     def _build(self) -> None:
        ...         self._add(0, "Linear")
        ...         self._add(1, "Cyclic")
        >>> class BType(u.AStructType):
        ...     def _build(self) -> None:
        ...         self._add("foo", AType())
        ...         self._add("mode", MType())
        ...         self._add("bar", u.ArrayType(AType(), 3), u.BWD)

    These types are automatically resolved by iterating over the Constant:

        >>> const = u.Const(BType(), "const_p")
        >>> for item in const:
        ...     print(repr(item))
        Const(BType(), 'const_p')
        Const(AType(), 'const_foo_p')
        Const(BitType(), 'const_foo_req_p')
        Const(ArrayType(UintType(16), 5), 'const_foo_data_p')
        Const(BitType(), 'const_foo_ack_p')
        Const(MType(), 'const_mode_p')
        Const(ArrayType(AType(), 3), 'const_bar_p')
        Const(ArrayType(BitType(), 3), 'const_bar_ack_p')
        Const(ArrayType(ArrayType(UintType(16), 5), 3), 'const_bar_data_p')
        Const(ArrayType(BitType(), 3), 'const_bar_req_p')
"""

from .ident import Ident


class Const(Ident):
    """
    Module Constant.

    Args:
        type_: Type.
        name: Name.

    Attributes:
        direction: Direction.
        doc: Documentation Container
        ifdef: IFDEF encapsulation


    ??? Example "Constant Examples"
        Basics:

            >>> import ucdp as u
            >>> cnt = u.Const(u.UintType(6), "cnt_p")
            >>> cnt
            Const(UintType(6), 'cnt_p')
            >>> cnt.type_
            UintType(6)
            >>> cnt.name
            'cnt_p'
            >>> cnt.basename
            'cnt'
            >>> cnt.suffix
            '_p'
            >>> cnt.doc
            Doc()

    ??? Example "Casting Constant"
        If the constant is casted via `int()` it returns `value` if set, other `type_.default`.

            >>> int(u.Const(u.UintType(6, default=2), "cnt_p"))
            2

    ??? Example "Constant are Singleton"
        Constant are Singleton:

            >>> u.Const(u.UintType(6), "cnt_p") is u.Const(u.UintType(6), "cnt_p")
            True
    """



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""File Set."""

from pathlib import Path

from .filelistparser import FileListParser
from .modbase import BaseMod
from .modfilelist import Paths, iter_modfilelists
from .object import LightObject, Object


class LibPath(LightObject):
    """
    Library and File Path.

    Attributes:
        libname: Library Name
        path: Path
    """

    libname: str
    path: Path


class FileSet(Object):
    """
    Module File List.

    Attributes:
        target: Target
        filepaths: Source Files
        inc_dirs: Include Directories
    """

    target: str | None = None
    filepaths: tuple[LibPath, ...]
    inc_dirs: tuple[Path, ...]

    @staticmethod
    def from_mod(
        topmod: BaseMod, name: str, target: str | None = None, filelistparser: FileListParser | None = None
    ) -> "FileSet":
        """Create ``FileSet` for ``mod``."""
        filepaths: list[LibPath] = []
        inc_dirs: list[Path] = []
        for mod, modfilelist in iter_modfilelists(topmod, name, target=target, filelistparser=filelistparser):
            libname = mod.libname
            _process(filepaths, inc_dirs, libname, modfilelist.dep_filepaths, modfilelist.dep_inc_dirs)  # type: ignore[arg-type]
            _process(filepaths, inc_dirs, libname, modfilelist.filepaths, modfilelist.inc_dirs)  # type: ignore[arg-type]

        return FileSet(target=target, filepaths=tuple(filepaths), inc_dirs=tuple(inc_dirs))

    def __iter__(self):
        for incdir in self.inc_dirs:
            yield f"-incdir {incdir}"
        for libfilepath in self.filepaths:
            yield str(libfilepath.path)


def _process(
    filepaths: list[LibPath],
    inc_dirs: list[Path],
    libname: str,
    add_filepaths: Paths | None,
    add_inc_dirs: Paths | None,
):
    # inc_dir
    for inc_dir in add_inc_dirs or []:
        if inc_dir not in inc_dirs:
            inc_dirs.append(inc_dir)
    # filepath
    for filepath in add_filepaths or []:
        libpath = LibPath(libname=libname, path=Path(filepath))
        if libpath not in filepaths:
            filepaths.append(libpath)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Cache.
"""

import os
from pathlib import Path
from shutil import rmtree

from anycache import AnyCache
from platformdirs import user_cache_path
from pydantic import BaseModel

from .logging import LOGGER


class Cache(BaseModel):
    """UCDP Caching System."""

    path: Path | None
    maxsize: int = 10 * 1024 * 1024

    @classmethod
    def init(cls) -> "Cache":
        """Initialize."""
        return Cache(path=get_cachepath())

    def disable(self):
        """Disable Caching."""
        self.path = None

    def clear(self):
        """Clear Cache."""
        if self.path:
            rmtree(self.path, ignore_errors=True)

    @property
    def templates_path(self) -> Path | None:
        """Path for Templates."""
        if not self.path:
            return None
        return self.path / "templates"

    def get_cache(self, name: str):
        """Create Cache Function Decorator."""
        return self._get_anycache(name).anycache

    def _get_anycache(self, name: str) -> AnyCache:
        if not self.path:
            return AnyCache(maxsize=0)
        return AnyCache(cachedir=self.path / name, maxsize=self.maxsize)


def get_cachepath() -> Path | None:
    """Get Cache Base Path."""
    try:
        envvar = os.environ["UCDP_CACHE"]
        if not envvar:
            return None
        path = Path(envvar)
    except KeyError:
        try:
            path = user_cache_path("ucdp", "iccode17", ensure_exists=True)
        except RuntimeError as exc:  # pragma: no cover
            LOGGER.warning(exc)
            return None
    try:
        path.mkdir(parents=True, exist_ok=True)
        (path / "loader").mkdir(exist_ok=True)
        (path / "templates").mkdir(exist_ok=True)
        (path / ".initialized").touch()
    except Exception as exc:
        LOGGER.warning(exc)
        return None
    return path


CACHE = Cache.init()



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Enumeration Types.

An enumeration is a normal base type with a specific mapping of values to a another meaning.

* :any:`EnumItem` - Enumeration item
* :any:`AEnumType` - Standard Enumeration
* :any:`AGlobalEnumType` - A public enumeration which fills up through all instances.
* :any:`DynamicEnumType` - A public enumeration which fills up per instance.
* :any:`EnaType` - Native single bit with `ena` and `dis` enumeration, active-high
* :any:`DisType` - Native single bit with `ena` and `dis` enumeration, low-high
"""

from typing import Any

from .object import Light
from .typebase import AScalarType
from .typebaseenum import BaseEnumType
from .typescalar import BitType, IntegerType


class AEnumType(BaseEnumType, Light):
    """
    Base class for all enumerations, behaves like a dictionary.

    Attributes:
        default: Default Value. Default value of `type` by default.

    The protected method `_build()` should be used to build the type.

    Definition of an enumeration:

    >>> import ucdp as u
    >>> class ModeType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(2, default=1)
    ...     def _build(self) -> None:
    ...         self._add(0, "linear", title="Linear Mode", descr="Just Linear", comment="be careful")
    ...         self._add(1, "cyclic", title="Cyclic Mode", descr="The Cyclic Mode")
    ...         self._add(2, "loop", title="Run in a Loop")

    Usage of an Enumeration:

    >>> mode = ModeType()
    >>> mode
    ModeType()

    The enumeration behaves like a `dict`, with elements hashed by `name`.
    But different to a regular `dict`, it returns items on pure iteration:

    >>> tuple(mode)
    (0, 1, 2)
    >>> mode.keys()
    dict_keys([0, 1, 2])
    >>> mode.values()
    dict_values([EnumItem(0, 'linear', doc=...), EnumItem(1, 'cyclic', doc=...), EnumItem(2, 'loop', doc=...)])
    >>> for key, item in mode.items():
    ...     print(key, item)
    0 EnumItem(0, 'linear', doc=...)
    1 EnumItem(1, 'cyclic', doc=...)
    2 EnumItem(2, 'loop', doc=...)

    Enumeration items have these attributes:

    >>> from tabulate import tabulate
    >>> print(tabulate([(item.key, item.value, item.doc) for item in mode.values()],
    ...                headers=(".key", ".value", ".doc")))
      .key  .value    .doc
    ------  --------  -------------------------------------------------------------------
         0  linear    Doc(title='Linear Mode', descr='Just Linear', comment='be careful')
         1  cyclic    Doc(title='Cyclic Mode', descr='The Cyclic Mode')
         2  loop      Doc(title='Run in a Loop')

    To retrieve an item by value:

    >>> mode.get_byvalue('loop')
    EnumItem(2, 'loop', doc=Doc(title='Run in a Loop'))
    >>> mode.get_byvalue('unknown')
    Traceback (most recent call last):
      ...
    ValueError: ModeType() does not contain value 'unknown'. Known values are 'linear', 'cyclic' and 'loop'.
    >>> mode.get_key('loop')
    2
    >>> mode.get_value(2)
    'loop'

    To check a value against the key, use the standard `check` method:

    >>> mode.check(0)
    0
    >>> mode.check(1)
    1

    To encode a mapped value, use the `encode` method

    >>> mode.encode('linear')
    0
    >>> mode.encode('cyclic')
    1
    >>> mode.encode('other')
    Traceback (most recent call last):
      ..
    ValueError: ModeType() does not contain value 'other'. Known values are 'linear', 'cyclic' and 'loop'.
    >>> mode.encode('other', usedefault=True)
    1

    Decoding works likewise:

    >>> mode.decode(0)
    'linear'
    >>> mode.decode(1)
    'cyclic'
    >>> mode.decode(3)
    Traceback (most recent call last):
      ...
    ValueError: ModeType() does not contain key 3. Known keys are 0, 1 and 2.
    >>> mode.decode(3, usedefault=True)
    'cyclic'

    You can also check, if a value is within the range:

    >>> 0 in mode
    True
    >>> 3 in mode
    False

    Enumerations are also singleton:

    >>> ModeType() is ModeType()
    True
    >>> ModeType() is ModeType(default=1)
    False

    >>> ModeType() == ModeType()
    True
    >>> ModeType() == ModeType(default=2)
    False

    Attributes `width`, `default` are taken from `keytype`:

    >>> mode = ModeType()
    >>> mode.width
    2
    >>> mode.default
    1
    >>> mode.check(3)
    3
    >>> mode.check(4)
    Traceback (most recent call last):
      ...
    ValueError: Value 4 is not a 2-bit integer with range [0, 3]

    Attributes `default` can be overwritten:

    >>> mode = ModeType(default=2)
    >>> mode.width
    2
    >>> mode.default
    2

    A mapping type, which translates one type to another:

    Definition of an enumeration:

    >>> import ucdp as u
    >>> class MappingType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(2)
    ...     valuetype: u.UintType = u.UintType(16)
    ...     def _build(self) -> None:
    ...         self._add(0, 7)
    ...         self._add(1, 31)
    ...         self._add(2, 2**14-1)
    ...         self._add(3, 2**16-1)
    >>> mapping = MappingType()
    >>> for item in mapping.values():
    ...     print(repr(item))
    EnumItem(0, 7)
    EnumItem(1, 31)
    EnumItem(2, 16383)
    EnumItem(3, 65535)

    Get Hex

    >>> mapping.get_hex()
    Hex('0x0')
    >>> mapping.get_hex(value=3)
    Hex('0x3')

    Size in Bits:

    >>> mapping.bits
    2

    To determine if the enumeration is fully decoded, use `is_full`:

    >>> import ucdp as u
    >>> class AType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(1)  # 2 possible values
    ...     def _build(self) -> None:
    ...         self._add(0, "linear")
    ...         self._add(1, "cyclic")
    >>> AType().is_full
    True

    >>> class BType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(2) # 4 possible values
    ...     def _build(self) -> None:
    ...         self._add(0, "linear")
    ...         self._add(1, "cyclic")
    >>> BType().is_full
    False

    Connections are only allowed to other :any:`EnumType` with the same key-value mapping.

    >>> class CType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(1)  # 2 possible values
    ...     def _build(self) -> None:
    ...         self._add(0, "linear", title="other comment")
    ...         self._add(1, "cyclic")

    >>> class DType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(1)  # 2 possible values
    ...     def _build(self) -> None:
    ...         self._add(0, "linear")

    >>> AType().is_connectable(AType())
    True
    >>> AType().is_connectable(BType())
    False
    >>> AType().is_connectable(CType())
    True
    >>> AType().is_connectable(DType())
    False

    Slicing:

    >>> BType()[1:0]
    UintType(2)
    >>> BType()[1]
    UintType(1)

    Values can be used twice:

    >>> class DuplType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(2)
    ...     def _build(self) -> None:
    ...         self._add(0, 'a')
    ...         self._add(1, 'b')
    ...         self._add(2, 'c')
    ...         self._add(3, 'b')

    >>> dupl = DuplType()
    >>> dupl
    DuplType()
    >>> dupl.get_value(1)
    'b'
    >>> dupl.get_value(3)
    'b'
    >>> dupl.get_key('b')
    1

    The `new()` method creates a new variant:

    >>> class MyType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(2)
    ...     def _build(self) -> None:
    ...         self._add(0, "linear")
    ...         self._add(1, "cyclic")
    ...         self._add(2, "auto")

    >>> MyType()
    MyType()
    >>> MyType().new(default=1)
    MyType(default=1)
    >>> MyType().new(filter_=lambda item: item.value != "cyclic")
    MyType()
    """

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        self._build()
        self._is_locked = True


class AGlobalEnumType(BaseEnumType, Light):
    """
    A singleton enumeration which can be filled outside `_build` and is **shared** between instances.

    >>> import ucdp as u
    >>> class CtrlType(u.AGlobalEnumType):
    ...     keytype: u.AScalarType = u.UintType(3)
    >>> ctrl = CtrlType()
    >>> ctrl.add(0, 'zero')

    >>> ctrl = CtrlType()
    >>> ctrl.add(1, 'one')
    >>> ctrl.add(7, 'seven')

    >>> ctrl.keys()
    dict_keys([0, 1, 7])

    >>> ctrl = CtrlType()
    >>> ctrl.keys()
    dict_keys([0, 1, 7])

    This is forbidden on normal enumeration:

    >>> class CtrlType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(3)
    ...     def _build(self) -> None:
    ...         pass
    >>> ctrl = CtrlType()
    >>> ctrl._add(0, 'zero')
    Traceback (most recent call last):
      ...
    ucdp.exceptions.LockError: CtrlType(): Cannot add item 0='zero'.
    """

    keytype: AScalarType = IntegerType()

    def add(self, key, value, title: str | None = None, descr: str | None = None, comment: str | None = None) -> None:
        """
        Add NamedObject To Enumeration.

        Attributes:
            key: key value to be mapped.
            value: Mapped value.
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
        """
        self._add(key=key, value=value, title=title, descr=descr, comment=comment)

    def _build(self) -> None:
        """Build Type."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        if self.__class__ is AGlobalEnumType:
            raise TypeError("Can't instantiate abstract class AGlobalEnumType. Please create a subclass.")
        self._build()


class DynamicEnumType(BaseEnumType):
    """
    A enumeration which can be filled outside `_build` and is **not** shared between instances.

    >>> import ucdp as u
    >>> class CtrlType(u.DynamicEnumType):
    ...     keytype: u.AScalarType = u.UintType(3)
    >>> ctrl = CtrlType()
    >>> ctrl.add(0, 'zero')
    >>> ctrl.keys()
    dict_keys([0])

    >>> ctrl = CtrlType()
    >>> ctrl.add(1, 'one')
    >>> ctrl.add(7, 'seven')
    >>> ctrl.keys()
    dict_keys([1, 7])

    This is forbidden on normal enumeration:

    >>> class CtrlType(u.AEnumType):
    ...     keytype: u.AScalarType = u.UintType(3)
    ...     def _build(self) -> None:
    ...         pass
    >>> ctrl = CtrlType()
    >>> ctrl._add(0, 'zero')
    Traceback (most recent call last):
      ...
    ucdp.exceptions.LockError: CtrlType(): Cannot add item 0='zero'.
    """

    keytype: AScalarType = IntegerType()

    def add(self, key, value, title: str | None = None, descr: str | None = None, comment: str | None = None) -> None:
        """
        Add NamedObject To Enumeration.

        Attributes:
            key: key value to be mapped.
            value: Mapped value.
            title: Full Spoken Name.
            descr: Documentation Description.
            comment: Source Code Comment.
        """
        self._add(key=key, value=value, title=title, descr=descr, comment=comment)

    def _build(self) -> None:
        """Build Type."""

    def model_post_init(self, __context: Any) -> None:
        """Run Build."""
        self._build()


class EnaType(AEnumType):
    """
    Enable (positive logic).

    >>> enable = EnaType()
    >>> enable
    EnaType()
    >>> enable.width
    1
    >>> enable.default
    0
    >>> for item in enable.values():
    ...     print(repr(item))
    EnumItem(0, 'dis', doc=Doc(title='disabled'))
    EnumItem(1, 'ena', doc=Doc(title='enabled'))

    >>> enable = EnaType(default=1)
    >>> enable.default
    1
    """

    keytype: AScalarType = BitType()
    title: str = "Enable"

    def _build(self) -> None:
        self._add(0, "dis", "disabled")
        self._add(1, "ena", "enabled")


class DisType(AEnumType):
    """
    Enable (positive logic).

    >>> disable = DisType()
    >>> disable
    DisType()
    >>> disable.width
    1
    >>> disable.default
    0
    >>> for item in disable.values():
    ...     print(repr(item))
    EnumItem(0, 'ena', doc=Doc(title='enabled'))
    EnumItem(1, 'dis', doc=Doc(title='disabled'))

    >>> disable = DisType(default=1)
    >>> disable.default
    1
    """

    keytype: AScalarType = BitType()
    title: str = "Disable"

    def _build(self) -> None:
        self._add(0, "ena", "enabled")
        self._add(1, "dis", "disabled")


class BusyType(AEnumType):
    """
    Busy.

    >>> busy = BusyType()
    >>> busy
    BusyType()
    >>> busy.width
    1
    >>> busy.default
    0
    >>> for item in busy.values():
    ...     print(repr(item))
    EnumItem(0, 'idle', doc=Doc(title='Idle'))
    EnumItem(1, 'busy', doc=Doc(title='Busy'))

    >>> busy = BusyType(default=1)
    >>> busy.default
    1
    """

    keytype: AScalarType = BitType()
    title: str = "Busy"

    def _build(self) -> None:
        self._add(0, "idle", "Idle")
        self._add(1, "busy", "Busy")
